/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg;

import static org.fortiss.af3.exploration.alg.ExplorationAlgActivator.getDefault;
import static org.fortiss.af3.exploration.alg.service.ExplorationServiceManager.getService;

import org.eclipse.ui.IStartup;
import org.fortiss.af3.exploration.alg.backend.Opt4JDseBackend;
import org.fortiss.af3.exploration.alg.feature.IExplorationFeature;
import org.fortiss.af3.exploration.alg.feature.partition.PartitionArchitectureExploration;
import org.fortiss.af3.exploration.alg.feature.platform.PlatformExploration;
import org.fortiss.af3.exploration.alg.feature.safetyarch.SafetyArchitectureExploration;
import org.fortiss.af3.exploration.alg.feature.scheduling.TTScheduleExploration;
import org.fortiss.af3.exploration.alg.feature.taskinstantiation.DesignDiversityExploration;
import org.fortiss.af3.exploration.alg.feature.taskmapping.TaskMappingExploration;
import org.fortiss.af3.exploration.alg.module.IEvaluatorModule;
import org.fortiss.af3.exploration.alg.module.common.evaluator.mapping.MappingEvaluatorConstraint;
import org.fortiss.af3.exploration.alg.module.common.evaluator.mapping.MappingEvaluatorObjective;
import org.fortiss.af3.exploration.alg.module.common.input.AllResourceInputModule;
import org.fortiss.af3.exploration.alg.module.common.input.DeployableTaskModule;
import org.fortiss.af3.exploration.alg.module.common.input.ExecutionUnitTargetModule;
import org.fortiss.af3.exploration.alg.module.common.input.GatewayUnitInputModule;
import org.fortiss.af3.exploration.alg.module.common.input.PlatformCommunicationGraphInputModule;
import org.fortiss.af3.exploration.alg.module.common.input.ResourceConnectionModule;
import org.fortiss.af3.exploration.alg.module.common.input.TaskGraphAcyclicInputModule;
import org.fortiss.af3.exploration.alg.module.common.input.TaskGraphInputModule;
import org.fortiss.af3.exploration.alg.module.common.input.TaskSignalInputModule;
import org.fortiss.af3.exploration.alg.module.common.input.TransmissionUnitInputModule;
import org.fortiss.af3.exploration.alg.module.expspectransf.constraint.AllocationPatternConverter;
import org.fortiss.af3.exploration.alg.module.expspectransf.dsmlext.PropertyToFixedValueLiteralConverter;
import org.fortiss.af3.exploration.alg.service.IExpSpecTransformationService;
import org.fortiss.af3.exploration.alg.service.IExplorationEvaluatationService;
import org.fortiss.af3.exploration.alg.service.IExplorationFeatureService;
import org.fortiss.af3.exploration.alg.service.IExplorationInputService;
import org.fortiss.af3.exploration.alg.service.IExplorationRepairService;
import org.fortiss.af3.exploration.alg.service.IExplorationService;
import org.fortiss.af3.exploration.alg.service.IExplorationTransformationService;
import org.fortiss.af3.exploration.alg.service.internal.ExpSpecTransformationService;
import org.fortiss.af3.exploration.alg.service.internal.ExplorationEvaluationService;
import org.fortiss.af3.exploration.alg.service.internal.ExplorationFeatureService;
import org.fortiss.af3.exploration.alg.service.internal.ExplorationInputService;
import org.fortiss.af3.exploration.alg.service.internal.ExplorationRepairService;
import org.fortiss.af3.exploration.alg.service.internal.transformation.ExplorationTransformationService;
import org.fortiss.af3.exploration.alg.service.module.IExpSpecTransformationModule;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.fortiss.af3.exploration.service.IDSEBackendService;
import org.fortiss.af3.exploration.smt.model.SMTConstraint;
import org.fortiss.af3.exploration.smt.model.SMTObjective;
import org.opt4j.core.DoubleValue;
import org.osgi.framework.BundleContext;

/**
 * Centralized class that provides the startup routines of all services provided by this plugin.
 * 
 * The order of execution is as follows:
 * - Initialize own services.
 * - Register with the exploration plugin. (Make the "Parent" aware of me)
 * - Start own services.
 * 
 * @author diewald
 */
public final class ExplorationAlg implements IStartup {

	/** {@inheritDoc} */
	@Override
	public void earlyStartup() {
		initialize();
		registerBackend();
		start();
	}

	/**
	 * Initializes the {@link IExplorationService}s contained in this plugin (first service
	 * activation stage).
	 */
	public static void initialize() {
		registerServices();

		getService(IExplorationInputService.class).initializeService();
		getService(IExplorationFeatureService.class).initializeService();
		getService(IExpSpecTransformationService.class).initializeService();
		getService(IExplorationEvaluatationService.class).initializeService();
		getService(IExplorationRepairService.class).initializeService();
		getService(IExplorationTransformationService.class).initializeService();

		// Registers the modules providing inputs from system models.
		registerInputModules();

		// Initializes/Registers the bundled exploration modules.
		registerExplorationFeatures();

		// Registers the ExplorationTarget / Exploration Specification Transformators
		registerExpSpecTransformators();

		// Add addtional Evaluators, not specific for a particular exploration feature.
		registerEvaluators();
	}

	/** Registers the provided services with the {@link BundleContext}. */
	private static void registerServices() {
		BundleContext context = getDefault().getBundle().getBundleContext();

		context.registerService(IExplorationInputService.class.getName(),
				ExplorationInputService.getInstance(), null);
		context.registerService(IExplorationFeatureService.class.getName(),
				ExplorationFeatureService.getInstance(), null);
		context.registerService(IExpSpecTransformationService.class.getName(),
				ExpSpecTransformationService.getInstance(), null);
		context.registerService(IExplorationEvaluatationService.class.getName(),
				ExplorationEvaluationService.getInstance(), null);
		context.registerService(IExplorationRepairService.class.getName(),
				ExplorationRepairService.getInstance(), null);
		context.registerService(IExplorationTransformationService.class.getName(),
				ExplorationTransformationService.getInstance(), null);
	}

	/** Registers the modules providing inputs from system models. */
	private static void registerInputModules() {
		IExplorationInputService iService = getService(IExplorationInputService.class);
		iService.registerExplorationInput(new ExecutionUnitTargetModule());
		iService.registerExplorationInput(new DeployableTaskModule());
		iService.registerExplorationInput(new ResourceConnectionModule());
		iService.registerExplorationInput(new PlatformCommunicationGraphInputModule());
		iService.registerExplorationInput(new TaskGraphInputModule());
		iService.registerExplorationInput(new TaskGraphAcyclicInputModule());
		iService.registerExplorationInput(new TaskSignalInputModule());
		iService.registerExplorationInput(new TransmissionUnitInputModule());
		iService.registerExplorationInput(new GatewayUnitInputModule());
		iService.registerExplorationInput(new AllResourceInputModule());
	}

	/** Registers the exploration modules bundled with this plugin. */
	private static void registerExplorationFeatures() {
		IExplorationFeatureService fService = getService(IExplorationFeatureService.class);
		fService.registerExplorationFeature(new SafetyArchitectureExploration());
		fService.registerExplorationFeature(new PartitionArchitectureExploration());
		fService.registerExplorationFeature(new PlatformExploration());
		fService.registerExplorationFeature(new TaskMappingExploration());
		fService.registerExplorationFeature(new DesignDiversityExploration());
		fService.registerExplorationFeature(new TTScheduleExploration());
	}

	/**
	 * Registers the locally defined {@link IExpSpecTransformationModule}s with the
	 * {@link IExpSpecTransformationService}.
	 */
	private static void registerExpSpecTransformators() {
		IExpSpecTransformationService tService = getService(IExpSpecTransformationService.class);
		// Transform DSML expressions to a more efficient internal representation.
		tService.registerModule(new PropertyToFixedValueLiteralConverter());
		// Transform DSML constraints to a more efficient internal representation.
		tService.registerModule(new AllocationPatternConverter());
	}

	/**
	 * Registers {@link IEvaluatorModule}s that are not specific for some
	 * {@link IExplorationFeature}.
	 */
	@SuppressWarnings("unchecked")
	private static void registerEvaluators() {
		IExplorationEvaluatationService eService =
				getService(IExplorationEvaluatationService.class);
		eService.registerEvaluator(new MappingEvaluatorObjective(
				(Class<? extends ExplorationTarget<DoubleValue>>)SMTObjective.class), null);
		eService.registerEvaluator(new MappingEvaluatorConstraint(SMTConstraint.class), null);
	}

	// TODO(INTEGRATION) Perform registration in
	// org.fortiss.af3.rcp.application.advisors.ApplicationWorkbenchAdvisor (or rather an AF3DSE
	// module) once this plugin has been integrated into the AF3 product.
	/** Register the DSE backend with the exploration plugin. */
	private static void registerBackend() {
		IDSEBackendService.getInstance().registerDSEBackend(new Opt4JDseBackend());
	}

	/**
	 * Starts the {@link IExplorationService}s contained in this plugin (second service activation
	 * stage).
	 */
	public static void start() {
		getService(IExplorationInputService.class).startService();
		getService(IExplorationFeatureService.class).startService();
		getService(IExpSpecTransformationService.class).startService();
		// The Transformation Service must be started before the Evaluation Service such that it can
		// check for the availability of required transformation modules.
		getService(IExplorationTransformationService.class).startService();
		getService(IExplorationEvaluatationService.class).startService();
		getService(IExplorationRepairService.class).startService();
	}
}
