/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend;

import static java.lang.System.currentTimeMillis;
import static org.fortiss.af3.exploration.alg.service.ExplorationServiceManager.getService;
import static org.fortiss.af3.exploration.alg.util.ExplorationAlgUtils.getAnnotatedMethod;

import java.util.ArrayList;
import java.util.Collection;

import org.eclipse.core.runtime.IProgressMonitor;
import org.fortiss.af3.exploration.alg.dse.RandomExploration;
import org.fortiss.af3.exploration.alg.dse.RandomExploration.RandomnessType;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.evaluate.EvaluatorInjectorProvider;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.evaluate.ValueProvider;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.evaluate.annotations.Validates;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.phenotype.ProvidingPhenotype;
import org.fortiss.af3.exploration.alg.exception.ExplorationException;
import org.fortiss.af3.exploration.alg.exception.ExplorationRuntimeException;
import org.fortiss.af3.exploration.alg.feature.IExplorationFeature;
import org.fortiss.af3.exploration.alg.module.EvaluatorModule;
import org.fortiss.af3.exploration.alg.service.IExplorationEvaluatationService;
import org.fortiss.af3.exploration.alg.service.IExplorationFeatureService;
import org.fortiss.af3.exploration.alg.service.module.IExplorationTargetEvaluator;
import org.fortiss.af3.exploration.model.ExplorationConstraint;
import org.fortiss.af3.exploration.model.ExplorationSpecification;
import org.fortiss.af3.exploration.moea.model.DseSpecification;
import org.fortiss.af3.exploration.moea.model.parameters.IOptimizer;
import org.opt4j.core.Value;

import com.google.inject.AbstractModule;
import com.google.inject.Injector;
import com.google.inject.Key;
import com.google.inject.Module;

/**
 * Base class for general optimization backends.
 * 
 * @author barner
 */
public abstract class ExplorerBackendBase<O extends IOptimizer> extends AbstractModule
		implements IExplorerBackend<O> {

	/** DSE specification (optimizer parameters, goal specification, input model, ...) */
	protected DseSpecification dse;

	/** Contains all ExplorationTargets of the exploration. */
	protected ExplorationSpecification expSpec;

	/** Representation of system model (=input) that is evaluated during this DSE run */
	protected Injector dseInjector;

	/**
	 * {@link IProgressMonitor} used to visualize the progress of the exploration process (may be
	 * {@code null}).
	 */
	protected IProgressMonitor progressMonitor;

	/**
	 * Constructor that initializes the DSE backend with the given optimizer independent parameters
	 * and the goal specification
	 */
	protected ExplorerBackendBase(DseSpecification dse, ExplorationSpecification expSpec,
			Injector dseInjector, IProgressMonitor progressMonitor) {
		this.dse = dse;
		this.expSpec = expSpec;
		this.dseInjector = dseInjector;
		this.progressMonitor = progressMonitor;

		// (Re-)Initialize the Random number generator
		initCentralRandomGenerator();
	}

	/**
	 * (Re-)Initializes the central random number generator of the DSE. It must be re-initialized on
	 * successive runs to obtain deterministic behavior.
	 */
	private void initCentralRandomGenerator() {
		if(!RandomExploration.getInstance().isInitialized()) {
			try {
				RandomExploration.getInstance().init(RandomnessType.realRandom,
						currentTimeMillis());
			} catch(Exception e) {
				// This is the correct place to initialize the central random generator.
				assert (false) : "The centralized random number generator of the DSE has been initialized in another location, where it should not be initialized.";
			}
		} else {
			try {
				RandomExploration.getInstance().reInit();
			} catch(Exception e) {
				// This is the correct place to re-initialize the central random generator, it has
				// been initialized in this place before.
				assert (false) : "The centralized random number generator of the DSE has not been initialized.";
			}
		}
	}

	/**
	 * Performs the initialization of the optimizer backend. Must be called from constructor in
	 * concrete derived classes.
	 */
	protected abstract void init(O optimizer) throws ExplorationException;

	/**
	 * Validation wrapper: It calls the validate methods from the
	 * {@link IExplorationTargetEvaluator}s.
	 */
	protected boolean validateSolution(ProvidingPhenotype providingPhenotype,
			ExplorationConstraint<?> target, Value<?> value) {
		Collection<IExplorationFeature> features = getService(IExplorationFeatureService.class)
				.getExplorationFeatures(dse.getFeatures());
		Collection<Class<? extends Module>> overrides = new ArrayList<>();
		features.forEach(f -> overrides.addAll(f.getModuleOverrides()));

		EvaluatorModule<?, ?, ?> evalModule =
				(EvaluatorModule<?, ?, ?>)getService(IExplorationEvaluatationService.class)
						.getEvaluatorInstance(target, overrides, features);

		Injector evaluationInjector =
				dseInjector.createChildInjector(providingPhenotype.getGenotypeProviderModule(),
						providingPhenotype.getPhenotypeProviderModule(), evalModule,
						new ValueProvider<>(value));
		EvaluatorInjectorProvider evalInjProvider =
				evaluationInjector.getInstance(EvaluatorInjectorProvider.class);
		evalInjProvider.setInjector(evaluationInjector);

		try {
			getAnnotatedMethod(evalModule.getClass(), Validates.class);
		} catch(ExplorationException e) {
			throw new ExplorationRuntimeException(
					"Failed to validate " + evalModule.getClass().getSimpleName() + ".", e);
		}

		return evaluationInjector.getInstance(Key.get(Boolean.class, Validates.class));
	}
}
