/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.constraint;

import static org.fortiss.tooling.common.util.LambdaUtils.filterByType;

import java.util.Collection;

import org.apache.commons.lang3.tuple.Pair;
import org.fortiss.af3.exploration.model.ExplorationSpecification;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.fortiss.af3.exploration.moea.model.constraints.InternalConstraint;
import org.opt4j.core.problem.Decoder;
import org.opt4j.core.problem.Evaluator;

import com.google.common.collect.HashMultimap;
import com.google.common.collect.Multimap;
import com.google.inject.TypeLiteral;

/**
 * Base class that holds the constraints that are specific for some artifact produced by the DSE. It
 * shall be used by operators and decoders to ensure the construction of valid artifacts. Only
 * "simple" constraints or ones produced by {@link Decoder}s shall be added to this map. For complex
 * constraints, the use of {@link Evaluator}s is advised.
 * 
 * @author diewald
 */
public class ConstrainedArtifactMap implements IConstrainedArtifactMap {

	/** Associates artifacts with constraints that a producer or modifier has to consider. */
	private Multimap<Pair<TypeLiteral<?>, TypeLiteral<?>>, InternalConstraint<?>> constrainedArtifacts =
			HashMultimap.create();

	/** Constructor. */
	public ConstrainedArtifactMap(ExplorationSpecification expSpec) {
		Collection<ExplorationTarget<?>> targets = expSpec.getTargets();
		for(InternalConstraint<?> constr : filterByType(targets, InternalConstraint.class)) {
			// TODO(AD): Does not yet handle multi-layer contraints.
			TypeLiteral<?> srcType = constr.getSourceType();
			TypeLiteral<?> tgtType = constr.getTargetType();
			constrainedArtifacts.put(Pair.of(srcType, tgtType), constr);
		}
	}

	/** {@inheritDoc} */
	@Override
	public Collection<InternalConstraint<?>> getConstraintsOf(TypeLiteral<?> srcType,
			TypeLiteral<?> tgtType) {
		return constrainedArtifacts.get(Pair.of(srcType, tgtType));
	}

	/** {@inheritDoc} */
	@Override
	public void addConstraint(InternalConstraint<?> constraint) {
		constrainedArtifacts.put(Pair.of(constraint.getSourceType(), constraint.getTargetType()),
				constraint);
	}

	/** {@inheritDoc} */
	@Override
	public void addConstraints(Collection<InternalConstraint<?>> constraints) {
		constraints.forEach(c -> addConstraint(c));
	}

	/** {@inheritDoc} */
	@Override
	public void removeArtifact(TypeLiteral<?> srcType, TypeLiteral<?> tgtType) {
		constrainedArtifacts.removeAll(Pair.of(srcType, tgtType));
	}
}
