/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.decode;

import static org.fortiss.tooling.common.util.LambdaUtils.getFirst;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import org.aopalliance.intercept.MethodInterceptor;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.Opt4JDseTaskProvider;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.DseInjectorProvider;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.EncodingProviderModule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.InfeasibleEncodingException;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.evaluate.ValueProvider;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.genotype.ProvidingGenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.phenotype.PhenotypeProviderGenerator;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.phenotype.ProvidingPhenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.thread.ThreadedMethod;
import org.fortiss.af3.exploration.alg.module.CreatorModule;
import org.fortiss.af3.exploration.alg.module.DecoderModule;
import org.fortiss.af3.exploration.alg.module.EvaluatorModule;
import org.fortiss.af3.exploration.alg.module.common.evaluator.BooleanValue;
import org.opt4j.core.Genotype;
import org.opt4j.core.Value;
import org.opt4j.core.problem.Decoder;
import org.opt4j.core.start.Opt4JTask;

import com.google.inject.AbstractModule;
import com.google.inject.Inject;
import com.google.inject.Injector;
import com.google.inject.Key;
import com.google.inject.ProvisionException;
import com.google.inject.TypeLiteral;

/**
 * {@link Decoder} that handles dependencies between {@link Decoder}s, {@link Genotype}s, and
 * {@link Phenotype}s using the Guice framework.
 * 
 * @author diewald
 */
public class GuiceDecoder extends AbstractModule
		implements Decoder<ProvidingGenotype, ProvidingPhenotype> {

	/** Defines the desired exploration result type. */
	// TODO: Remove this hard coded line and get the desired result from the backend interface.
	protected Set<TypeLiteral<? extends Phenotype>> resultTypes;

	/** Set of {@link CreatorModule}s provided by the enabled exploration features. */
	private Collection<TypeLiteral<? extends DecoderModule<?>>> enabledDecoders;

	/** {@link Opt4JTask} provider used to obtain the opt4j {@link Injector}. */
	@Inject
	protected Opt4JDseTaskProvider opt4JProvider;

	/** Child {@link Injector} for the dependency management of the decode phase. */
	protected Injector decodingInjector;

	/** References the top-most injector of the DSE. */
	private Injector dseInjector;

	/**
	 * Holds the {@link MethodInterceptor} that is executed when a decode method capable of
	 * launching sub-decoders is hit.
	 */
	protected DecodeInterceptor decodingInterceptor;

	/** Constructor. Remembers the demanded {@link Phenotype}ss to be produced. */
	@Inject
	public GuiceDecoder(Set<TypeLiteral<? extends Phenotype>> resultTypes,
			DseInjectorProvider dseInjectorProvider) {
		this.resultTypes = resultTypes;
		this.dseInjector = dseInjectorProvider.getInjector();
	}

	// TODO: Pass/calculate the set of enabled decoders with the constructor.
	/**
	 * Defines the decoders that are required in this DSE execution to produce the demanded results.
	 */
	public void setEnabledDecoders(
			Collection<TypeLiteral<? extends DecoderModule<?>>> enabledDecoders) {
		this.enabledDecoders = enabledDecoders;
	}

	/** {@inheritDoc} */
	@Override
	@ThreadedMethod
	public ProvidingPhenotype decode(ProvidingGenotype genotype) {
		boolean isInfeasible = genotype.isInfeasible();

		Injector decodingInjector = createDecodingInjector(genotype);

		// FIXME: Hack to get the correct instance: Is multiprovisioning supported by Guice?
		PhenotypeCache aggregatePhenotypes = decodingInjector.getInstance(PhenotypeCache.class);
		aggregatePhenotypes.addAllPreservedPhenos(genotype);
		if(!isInfeasible) {
			try {
				DecodeInterceptor interceptor =
						decodingInjector.getInstance(DecodeInterceptor.class);
				interceptor.setInterceptParameters(decodingInjector, genotype.getConstraintMap(),
						Phenotype.class, PhenotypeCache.class);

				TypeLiteral<? extends Phenotype> resultType = getFirst(resultTypes).get();
				decodingInjector.getInstance(Key.get(resultType));
			} catch(ProvisionException e) {
				if(e.getCause() instanceof InfeasibleEncodingException) {
					isInfeasible = true;
				} else {
					throw e;
				}
			}
		}
		// The decoding Injector creates the Phenotype instances. They are cached in the
		// interceptor after each of the decode methods returns.

		return createProvidingPhenotype(genotype, decodingInjector, aggregatePhenotypes,
				isInfeasible);
	}

	/**
	 * Constructs a {@link ProvidingGenotype} from the input {@link ProvidingGenotype} and the
	 * decoded {@link Phenotype}s such that it can be consumed by {@link EvaluatorModule}s.
	 */
	private ProvidingPhenotype createProvidingPhenotype(ProvidingGenotype genotype,
			Injector decodingInjector, PhenotypeCache aggregatePhenotypes, boolean isInfeasible) {
		PhenotypeProviderGenerator phenoProviderGenerator =
				decodingInjector.getInstance(PhenotypeProviderGenerator.class);
		EncodingProviderModule<Phenotype> phenotProviderModule =
				phenoProviderGenerator.generateInstanceProviders(aggregatePhenotypes.entrySet());

		Map<TypeLiteral<? extends Phenotype>, Phenotype> solutionHolder = new HashMap<>();
		solutionHolder.putAll(aggregatePhenotypes);
		ProvidingPhenotype providingPhenotype = new ProvidingPhenotype(genotype.getProviderModule(),
				phenotProviderModule, solutionHolder, genotype);
		if(isInfeasible) {
			providingPhenotype.setInfeasible();
		}

		return providingPhenotype;
	}

	/**
	 * Creates a decoding {@link Injector} as child {@link Injector} if the DSE-wide
	 * {@link Injector}. Thread-safe.
	 */
	private synchronized Injector createDecodingInjector(ProvidingGenotype genotype) {
		return opt4JProvider.getInjector().createChildInjector(this, genotype.getProviderModule());
	}

	/** {@inheritDoc} */
	@Override
	protected void configure() {
		install(this);
		install(PhenotypeCache.getPrototype());
		// Install a dummy value provider for the validate methods.
		install(new ValueProvider<Value<?>>(new BooleanValue(true)));

		for(TypeLiteral<? extends DecoderModule<?>> module : enabledDecoders) {
			install(dseInjector.getInstance(Key.get(module)));
		}
	}
}
