/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.sysmodel.arch;

import static java.util.Collections.emptyMap;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Map;
import java.util.Set;

import org.fortiss.af3.component.model.Component;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceConnectionAdapter.ConnectionType;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.AcyclicTaskGraph;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.DefaultTaskGraph;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.ITaskGraph;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.TaskGraph;
import org.fortiss.af3.exploration.alg.feature.platform.PlatformCommunicationGraphEncoding;
import org.fortiss.af3.platform.model.ExecutionUnit;
import org.fortiss.af3.platform.model.IPlatformResource;
import org.fortiss.af3.platform.model.PlatformArchitecture;
import org.fortiss.tooling.base.model.element.IConnector;

import com.google.common.collect.Multimap;
import com.google.inject.AbstractModule;
import com.google.inject.Provides;
import com.google.inject.name.Named;

/**
 * Adapter to the system model used as input for the DSE which is created for a specific DSE run.
 * 
 * @author barner, huang
 */
public class SystemModelAdapter<C, E, SO, SI, G, TR, M> extends AbstractModule {

	/** Deployable software components, to be mapped to deployment targets provided by platform. */
	protected Collection<ITaskAdapter> tasks;

	/** Collection of deployable components from the base model. */
	protected Collection<C> taskObjects;

	/** Collection of signals exchanged between tasks. */
	protected Collection<ISignalAdapter> signals;

	/**
	 * Look up table to determine the input/output directions of {@link IConnector} types of the
	 * target {@link PlatformArchitecture}.
	 */
	protected Map<Class<? extends IConnector>, ConnectionType> connectionTypeLUT = emptyMap();

	/**
	 * Deployable software components that may be replicated (per default: all @{code
	 * deployableComponent}s.
	 */
	protected Collection<ITaskAdapter> nonSafetyFunctionTasks;

	/**
	 * Collection of software components that will be actually executed. This includes realizations
	 * of "abstract" software components (-> interfaces) that are contained in the
	 * {@code deployableComponents} collection.
	 */
	protected Multimap<ITaskAdapter, ITaskAdapter> taskInstances;

	/** All {@link ExecutionUnit}s that are present in the target platform. */
	protected Collection<IExecutionUnitAdapter> allPresentExecutionUnits;

	/** {@link Collection} of memories present in the system. */
	protected Set<IMemoryUnitAdapter> memoryUnits;

	/**
	 * Gateway units that are employed to manage the communication between multiple networks
	 * (connecting the same or different layers).
	 */
	protected Set<IGatewayUnitAdapter> gatewayUnits;

	/** TDMA arbitrated transmission unit */
	protected Set<ITransmissionUnitAdapter> transmissionUnits;

	/** Contains all available communication resources of the platform. */
	protected Collection<ICommunicationResourceAdapter> availableCommunicationResource;

	/** Contains all resources that are available in the given platform. */
	protected Set<IResourceAdapter> availableResources;

	/** Contains the connections between the {@link IResourceAdapter}s of the platform. */
	protected Set<IResourceConnectionAdapter> platformConnections;

	/**
	 * Collection of the safety functions adapters that represent the safety functions from the
	 * input models.
	 */
	protected Collection<ISafetyFunctionAdapter> safetyFunctionAdapters = new ArrayList<>();

	/** The {@link PlatformCommunicationGraphEncoding} that is given by the current platform. */
	protected PlatformCommunicationGraphEncoding platformCommunicationGraph;

	/** Hyper-period of independent tasks graphs */
	protected Double hyperPeriod;

	/**
	 * Map: deployable component (= task) -> corresponding (independent, connected)
	 * {@link DefaultTaskGraph}.
	 */
	protected Map<ITaskAdapter, TaskGraph> tasksInGraph;

	/**
	 * Map: deployable component (= task) -> corresponding (independent, connected)
	 * {@link AcyclicTaskGraph}.
	 */
	protected Map<ITaskAdapter, AcyclicTaskGraph> componentInAcyclicGraph;

	/** Defines the fault containment region defined for the target {@link PlatformArchitecture}. */
	protected Class<? extends IPlatformResource> faultContainmentRegion;

	/**
	 * Adds the given collection of {@link ISafetyFunctionAdapter}s to {@code this}
	 * {@link SystemModelAdapter} that represents the input system models. In contrast to other
	 * input parameters, the {@link ISafetyFunctionAdapter}s are created dynamically if a
	 * safety constraint is present which results in the safety functions to be
	 * extracted <b>after</b> the {@link SystemModelAdapter} has been constructed.
	 * <p>
	 * This method updates the set of replicable components by removing those components that are
	 * referenced by the safety function.
	 */
	// TODO: remove the enclosed cast. The difficulty is that the ISafetyFunctionAdapters are added
	// from within the DSE where the system types are not known any more.
	// It would be more appropriate to shift this addition before launching the DSE.
	@SuppressWarnings({"unchecked", "rawtypes"})
	public void
			addSafetyFunctionAdapters(Collection<ISafetyFunctionAdapter> safetyFunctionAdapters) {
		if(safetyFunctionAdapters == null) {
			return;
		}
		this.safetyFunctionAdapters.addAll(safetyFunctionAdapters);

		for(ISafetyFunctionAdapter sfAdapter : safetyFunctionAdapters) {
			nonSafetyFunctionTasks.removeAll(sfAdapter.getDiagnosticComponents());
			nonSafetyFunctionTasks.removeAll(sfAdapter.getRealizingComponents());
			nonSafetyFunctionTasks.remove(sfAdapter.getOutputComponent());
		}
	}

	/**
	 * Prints the given collection of {@link ITaskGraph}s to the standard output and displays them
	 * in
	 * a separate dialog.
	 */
	@SuppressWarnings("unused")
	private <T extends ITaskGraph<?>> void printAndDisplayTaskGraphs(Collection<T> taskGraphs) {
		for(T taskGraph : taskGraphs) {
			System.out.println(
					"Application " + taskGraph + ":" + " period = " + taskGraph.getHyperPeriod());
			System.out.println(taskGraph);
			System.out.println("");

			// Open a dialog for the visualization of the TaskGraph
			taskGraph.display();
		}
	}

	/**
	 * Returns the deployable (software) component from the original logical architecture that my be
	 * replicated. May contain "abstract" {@link Component}s.
	 */
	@Provides
	@Named("NonSafetyFunctionTasks")
	public Collection<ITaskAdapter> getNonSafetyFunctionTasks() {
		return nonSafetyFunctionTasks;
	}

	/**
	 * Returns the deployable (software) components that can be actually executed on
	 * {@link ExecutionUnit}s, i.e. software components that are backed by an implementation.
	 */
	@Provides
	public Multimap<ITaskAdapter, ITaskAdapter> getTaskInstanceMap() {
		return taskInstances;
	}

	/**
	 * Returns the {@link ISignalAdapter} corresponding the sender and receiver {@link Component}.
	 * Returns null if no channel exists between the two given components.
	 */
	public ISignalAdapter getSignalAdapter(ITaskAdapter source, ITaskAdapter target) {
		for(ISignalAdapter sig : signals) {
			if(sig.getSource() == source && sig.getTargets().contains(target)) {
				return sig;
			}
		}
		return null;
	}

	/**
	 * Returns the {@link ISafetyFunctionAdapter}s constructed from the safety functions that are
	 * present in the input models.
	 */
	@SuppressWarnings("unchecked")
	@Provides
	public Collection<ISafetyFunctionAdapter> getSafetyFunctionAdapters() {
		if(safetyFunctionAdapters != null) {
			return safetyFunctionAdapters;
		}
		return Collections.emptySet();
	}

	/** Returns the fault containment region specified for the target platform. */
	public Class<? extends IPlatformResource> getFaultContainmentRegion() {
		return faultContainmentRegion;
	}
}
