/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.feature.partition;

import static java.util.Arrays.asList;

import org.fortiss.af3.exploration.alg.feature.ExplorationFeature;
import org.fortiss.af3.exploration.alg.feature.partition.evaluator.NoEmptyPartitionConstraintEvaluator;
import org.fortiss.af3.exploration.alg.feature.partition.evaluator.PartitionNumberEvaluator;
import org.fortiss.af3.exploration.alg.feature.partition.evaluator.SameSilInPartitionConstraintEvaluator;
import org.fortiss.af3.exploration.alg.feature.partition.evaluator.internal.TaskPartitionAssignmentConstraintEvaluator;
import org.fortiss.af3.exploration.alg.feature.partition.transformation.PartitionToExecutionUnitTransformation;
import org.fortiss.af3.exploration.alg.feature.partition.transformation.PartitionTransformation;
import org.fortiss.af3.exploration.alg.feature.partition.transformation.TaskToPartitionTransformation;
import org.fortiss.af3.exploration.alg.feature.partition.variable.Partition;
import org.fortiss.af3.exploration.alg.feature.partition.variable.PartitionMappingCreator;
import org.fortiss.af3.exploration.alg.feature.partition.variable.PartitionMappingEncoding;
import org.fortiss.af3.exploration.alg.feature.partition.variable.TaskPartitionMappingCreator;
import org.fortiss.af3.exploration.alg.feature.partition.variable.TaskPartitionMappingEncoding;
import org.fortiss.af3.exploration.alg.feature.partition.variable.operators.MutateTaskPartitionMappingModule;
import org.fortiss.af3.exploration.alg.feature.partition.variable.operators.PartitionMappingModule;
import org.fortiss.af3.exploration.alg.feature.platform.PlatformCommunicationGraphEncoding;
import org.fortiss.af3.exploration.alg.feature.platform.problem.PlatformCommunicationGraphExecUnitInstanceDecoder;
import org.fortiss.af3.exploration.alg.feature.taskmapping.problem.TaskExUMappingIdentityDecoder;
import org.fortiss.af3.exploration.moea.model.constraints.AssignmentConstraint;
import org.fortiss.af3.exploration.moea.model.constraints.NoEmptyPartitionConstraint;
import org.fortiss.af3.exploration.moea.model.constraints.SameSilInPartitionConstraint;
import org.fortiss.af3.exploration.moea.model.objectives.PartitionMinObjective;

import com.google.inject.Injector;
import com.google.inject.TypeLiteral;

/**
 * Adds a virtualization layer to the mapping problem such that an optimal number of
 * {@link Partition}s is found that provide time/space separation.
 * 
 * @author diewald
 */
public class PartitionArchitectureExploration extends ExplorationFeature {

	/** Constructor. */
	public PartitionArchitectureExploration() {
		super("Partition Architecture Exploration");

		addVariable(TypeLiteral.get(PartitionMappingEncoding.class), new PartitionMappingCreator(),
				asList(new PartitionMappingModule()));
		addVariable(TypeLiteral.get(TaskPartitionMappingEncoding.class),
				new TaskPartitionMappingCreator(), asList(new MutateTaskPartitionMappingModule()));

		addProblem(TypeLiteral.get(PlatformCommunicationGraphEncoding.class),
				new PartitionMappingDecoderGraph());
		addProblem(TypeLiteral.get(TaskPartitionMappingEncoding.class),
				new TaskPartitionMappingIdentityDecoder());
		addModuleOverride(TaskExUMappingIdentityDecoder.class);
		addModuleOverride(PlatformCommunicationGraphExecUnitInstanceDecoder.class);

		addEvaluator(new NoEmptyPartitionConstraintEvaluator(NoEmptyPartitionConstraint.class));
		addEvaluator(new SameSilInPartitionConstraintEvaluator(SameSilInPartitionConstraint.class));
		addEvaluator(new PartitionNumberEvaluator(PartitionMinObjective.class));
		addEvaluator(new TaskPartitionAssignmentConstraintEvaluator(AssignmentConstraint.class));

		addTransformator(new PartitionTransformation());
		addTransformator(new PartitionToExecutionUnitTransformation());
		addTransformator(new TaskToPartitionTransformation());
	}

	/** {@inheritDoc} */
	@Override
	public boolean isApplicableToInputs(Injector dseInjector) {
		return true;
	}
}
