/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.feature.platform;

import static org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceConnectionAdapter.ConnectionType.INCOMING;
import static org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceConnectionAdapter.ConnectionType.OUTGOING;
import static org.fortiss.tooling.common.util.LambdaUtils.getFirst;

import java.util.Collection;
import java.util.Set;
import java.util.function.Predicate;

import org.eclipse.draw2d.graph.DirectedGraph;
import org.fortiss.af3.exploration.alg.display.JGraphTVisualizer;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.phenotype.PhenotypeBase;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IGatewayUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceConnectionAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITransmissionUnitAdapter;
import org.fortiss.af3.exploration.alg.service.module.IExplorationEncoding;
import org.fortiss.af3.exploration.alg.util.GraphUtils;
import org.jgrapht.Graphs;
import org.jgrapht.graph.DefaultDirectedGraph;
import org.jgrapht.graph.DefaultEdge;

import com.google.inject.Inject;

/**
 * Abstracts the communication network present in a platform system meta-model to a flat graph. This
 * graph is needed to manage the communication of logical {@link ITaskAdapter}s after
 * they have been deployed onto a concrete platform. More precisely, the graph is used for routing
 * of (physical) messages and their scheduling on the communication resources, like busses.
 * 
 * @author diewald
 */
public class PlatformCommunicationGraphEncoding extends PhenotypeBase
		implements IExplorationEncoding {
	/** The graph representing the communication structure of a platform. */
	protected DefaultDirectedGraph<IResourceAdapter, DefaultEdge> platformGraph;

	/** Constructor for the initial version of the Graph w/o available Partitions. */
	@Inject
	public PlatformCommunicationGraphEncoding(Set<IExecutionUnitAdapter> execUnits,
			Set<IGatewayUnitAdapter> gatewayUnits,
			Set<ITransmissionUnitAdapter> transmissionUnits,
			Set<IResourceConnectionAdapter> resourceConnections) {
		this.platformGraph =
				createGraph(execUnits, gatewayUnits, transmissionUnits, resourceConnections);
	}

	/** Copy constructor. */
	@SuppressWarnings("unchecked")
	public PlatformCommunicationGraphEncoding(PlatformCommunicationGraphEncoding pcgEnc) {
		this.platformGraph =
				(DefaultDirectedGraph<IResourceAdapter, DefaultEdge>)pcgEnc.platformGraph
						.clone();
	}

	/**
	 * Creates the underlying platform graph based on the given HW elements of the SAME platform
	 * model.
	 */
	public DefaultDirectedGraph<IResourceAdapter, DefaultEdge> createGraph(
			Set<IExecutionUnitAdapter> execUnits, Set<IGatewayUnitAdapter> gatewayUnits,
			Set<ITransmissionUnitAdapter> transmissionUnits,
			Set<IResourceConnectionAdapter> resourceConnections) {
		DefaultDirectedGraph<IResourceAdapter, DefaultEdge> platformGraph =
				new DefaultDirectedGraph<IResourceAdapter, DefaultEdge>(DefaultEdge.class);

		// Add all resources present in the platform as vertices to the "connection graph"
		Graphs.addAllVertices(platformGraph, execUnits);
		Graphs.addAllVertices(platformGraph, gatewayUnits);
		Graphs.addAllVertices(platformGraph, transmissionUnits);

		// Successively connect the resources in the graph. The edges are given within the
		// collection resource connectors are assumed to be fully connected.
		for(IResourceConnectionAdapter connection : resourceConnections) {
			IResourceAdapter srcResource = connection.getSourceResource();
			// We add edges for each receiver in order to support broadcast connections
			for(IResourceAdapter currentReceiverResource : connection.getTargetResources()) {
				// If the current resource is a child of a {@link TransmissionUnit} (like a NoC
				// router), respect the direction of connections such that complex networks are
				// correctly reflected.
				if(connection.getConnectionType(currentReceiverResource) == OUTGOING) {
					platformGraph.addEdge(srcResource, currentReceiverResource);
				} else if(connection.getConnectionType(currentReceiverResource) == INCOMING) {
					platformGraph.addEdge(currentReceiverResource, srcResource);
				} else {
					// Add the edges of a single connection. We add to edges for each connection
					// such that an undirected graph is generated effectively. It is required to use
					// this "trick" in order to be able to use the standard shortest path
					// algorithms.
					platformGraph.addEdge(srcResource, currentReceiverResource);
					platformGraph.addEdge(currentReceiverResource, srcResource);
				}
			}
		}

		return platformGraph;
	}

	/**
	 * Returns the subgraph of the given {@link IResourceAdapter}. The subgraph is obtained by
	 * "cutting" the connection to directly connected {@link ITransmissionUnitAdapter}s.
	 */
	public DefaultDirectedGraph<IResourceAdapter, DefaultEdge>
			getSubTreeOf(Collection<IResourceAdapter> containedResources) {
		DefaultDirectedGraph<IResourceAdapter, DefaultEdge> subGraph =
				new DefaultDirectedGraph<>(DefaultEdge.class);
		Predicate<IResourceAdapter> isNOTContainedElement = v -> !containedResources.contains(v);
		IResourceAdapter containedResource = getFirst(containedResources).get();
		GraphUtils.getSubGraph(platformGraph, containedResource, isNOTContainedElement, subGraph);
		return subGraph;
	}

	/** Creates a Dialog that illustrates this {@link DirectedGraph}. */
	public void display() {
		JGraphTVisualizer<IResourceAdapter, DefaultEdge> taskViz =
				new JGraphTVisualizer<IResourceAdapter, DefaultEdge>(platformGraph);
		taskViz.displayGraph("Platform communication graph");
	}

	/**
	 * Returns the actual graph that contains information about the {@link IResourceAdapter}s and
	 * their connections.
	 */
	public DefaultDirectedGraph<IResourceAdapter, DefaultEdge> getActualGraph() {
		return platformGraph;
	}
}
