/*-------------------------------------------------------------------------+
| Copyright 2019 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.feature.platform.input;

import static org.fortiss.tooling.kernel.utils.EcoreUtils.getFirstParentWithType;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.af3.AF3ExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.mappers.ExecUnitMapper;
import org.fortiss.af3.exploration.alg.feature.platform.PlatformExecUnitTemplateEncoding;
import org.fortiss.af3.exploration.alg.module.InputModule;
import org.fortiss.af3.exploration.alg.module.InputProvider;
import org.fortiss.af3.exploration.dseml.model.expression.SuperSet;
import org.fortiss.af3.platform.hierarchic.model.tile.Tile;
import org.fortiss.af3.platform.model.ExecutionUnit;
import org.fortiss.tooling.base.model.element.IModelElement;

import com.google.common.collect.HashMultimap;
import com.google.common.collect.Multimap;
import com.google.inject.Provides;
import com.google.inject.Singleton;

/**
 * Maps "parent" {@link ExecutionUnit}s of a hierarchical platform to a contained template
 * {@link ExecutionUnit}.
 * 
 * @author diewald
 */
@Singleton
public class PlatformExecUnitParentTemplateMapping
		extends InputModule<PlatformExecUnitTemplateEncoding> {

	/** Provides the mapping of parents to the templates of their children to instantiate. */
	@InputProvider
	@Provides
	@Singleton
	public PlatformExecUnitTemplateEncoding getPlatformExecUnitTemplateEncoding(
			ExecUnitMapper execUnitMapper, SuperSet<ExecutionUnit> execUnits,
			Set<IResourceAdapter> allResources) {
		Map<IExecutionUnitAdapter, IExecutionUnitAdapter> parentToTemplateMap = new HashMap<>();

		// We are considering tiles as the replication targets, so we walk up from our input set to
		// only get those tiles whose children are defined as mapping targets.
		Set<Tile> parentTiles = new HashSet<>();
		for(ExecutionUnit execUnit : execUnits.getEntries()) {
			parentTiles.add(getFirstParentWithType(execUnit, Tile.class));
		}

		Map<Tile, IExecutionUnitAdapter> templateToAdapterMap = new HashMap<>();
		Collection<ExecutionUnit> processedParents = new HashSet<>();
		for(ExecutionUnit execUnit : parentTiles) {
			IExecutionUnitAdapter targetExecUnit = execUnitMapper.getAdapter(execUnit);
			// Create the template adapter if the template is not part of the allocation targets
			targetExecUnit = (targetExecUnit != null) ? targetExecUnit
					: new AF3ExecutionUnitAdapter(execUnit);
			templateToAdapterMap.put(targetExecUnit.getObject(Tile.class), targetExecUnit);
			ExecutionUnit parentExecUnit = (ExecutionUnit)execUnit.getContainer();
			if(processedParents.contains(parentExecUnit)) {
				throw new RuntimeException("The resource " + parentExecUnit.getName() +
						" contains more than one template execution unit. This is not supported" +
						" in the PlatformExploration feature");
			}
			processedParents.add(parentExecUnit);
			parentToTemplateMap.put(new AF3ExecutionUnitAdapter(parentExecUnit), targetExecUnit);
		}

		Multimap<IExecutionUnitAdapter, IResourceAdapter> templateContainedResources =
				HashMultimap.create();
		for(IResourceAdapter resource : allResources) {
			IModelElement af3Element = resource.getObject(IModelElement.class);
			Tile potentialTemplate = getFirstParentWithType(af3Element, Tile.class);
			if(potentialTemplate != null && parentTiles.contains(potentialTemplate)) {
				IExecutionUnitAdapter execUnitAdp = templateToAdapterMap.get(potentialTemplate);
				templateContainedResources.put(execUnitAdp, resource);
			}
		}

		return new PlatformExecUnitTemplateEncoding(parentToTemplateMap,
				templateContainedResources);
	}
}
