/*-------------------------------------------------------------------------+
| Copyright 2019 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.feature.scheduling;

import static java.util.Collections.emptyList;
import static java.util.stream.Collectors.toSet;
import static org.eclipse.core.runtime.Assert.isNotNull;
import static org.fortiss.af3.exploration.alg.util.ExplorationAlgUtils.copyExplorationSpecification;
import static org.fortiss.af3.exploration.smt.modeltransformation.BasicScheduleConstraint.generatePureScheduleConstraints;
import static org.fortiss.af3.exploration.smt.modeltransformation.ConstraintDefinitionUtils.createDurationLiteral;
import static org.fortiss.af3.exploration.smt.modeltransformation.ConstraintDefinitionUtils.createStartTimeLiteral;
import static org.fortiss.af3.exploration.smt.modeltransformation.ConstraintDefinitionUtils.generateForAll;
import static org.fortiss.af3.exploration.smt.util.AllocationToSchedule.generateSchedule;
import static org.fortiss.af3.exploration.smt.util.Z3ModelElementFactory.createSMTConstraint;
import static org.fortiss.af3.exploration.smt.util.Z3ModelElementFactory.createSMTObjective;
import static org.fortiss.af3.exploration.util.DSMLModelElementFactory.createArithmeticLiteral;
import static org.fortiss.af3.exploration.util.DSMLModelElementFactory.createBooleanLiteral;
import static org.fortiss.af3.exploration.util.DSMLModelElementFactory.createLessEqual;
import static org.fortiss.af3.exploration.util.DSMLModelElementFactory.createMinimize;
import static org.fortiss.af3.exploration.util.DSMLModelElementFactory.createPlus;
import static org.fortiss.af3.exploration.util.DSMLModelElementFactory.createSet;
import static org.fortiss.af3.exploration.util.DSMLModelElementFactory.createSum;
import static org.fortiss.af3.exploration.util.DSMLModelElementFactory.createSuperSet;
import static org.fortiss.af3.exploration.util.ExplorationModelElementFactory.createExplorationSolution;
import static org.fortiss.tooling.common.util.LambdaUtils.getFirst;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getChildrenWithType;

import java.math.BigDecimal;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;

import org.eclipse.core.runtime.Assert;
import org.eclipse.emf.ecore.EObject;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.decode.Decodes;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ICommunicationResourceAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IRequestAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.mappers.TaskMapper;
import org.fortiss.af3.exploration.alg.feature.platform.PlatformCommunicationGraphEncoding;
import org.fortiss.af3.exploration.alg.feature.scheduling.transform.MessageToRouteTransformation;
import org.fortiss.af3.exploration.alg.feature.taskmapping.TaskToExUTransformation;
import org.fortiss.af3.exploration.alg.feature.taskmapping.problem.comm.Message;
import org.fortiss.af3.exploration.alg.feature.taskmapping.problem.comm.MessageEncoding;
import org.fortiss.af3.exploration.alg.feature.taskmapping.variable.TaskExUMappingEncoding;
import org.fortiss.af3.exploration.alg.feature.taskmapping.variable.TaskExUMappingEntry;
import org.fortiss.af3.exploration.alg.module.DecoderModule;
import org.fortiss.af3.exploration.dseml.model.arithmetic.ModelElementPropertyLiteral;
import org.fortiss.af3.exploration.dseml.model.arithmetic.Plus;
import org.fortiss.af3.exploration.dseml.model.arithmetic.Sum;
import org.fortiss.af3.exploration.dseml.model.booleanp.BooleanLiteral;
import org.fortiss.af3.exploration.dseml.model.booleanp.ForAll;
import org.fortiss.af3.exploration.dseml.model.booleanp.comparison.LessEqual;
import org.fortiss.af3.exploration.dseml.model.expression.Set;
import org.fortiss.af3.exploration.dseml.model.expression.SuperSet;
import org.fortiss.af3.exploration.dseml.model.function.Minimize;
import org.fortiss.af3.exploration.model.ExplorationSpecification;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.fortiss.af3.exploration.model.SuperSetMap;
import org.fortiss.af3.exploration.model.solutions.ExplorationSolution;
import org.fortiss.af3.exploration.model.solutions.SolutionState;
import org.fortiss.af3.exploration.smt.model.SMTConstraint;
import org.fortiss.af3.exploration.smt.model.SMTObjective;
import org.fortiss.af3.exploration.smt.modeltransformation.DSMLtoSMTTransformator;
import org.fortiss.af3.exploration.smt.modeltransformation.ScheduleRun;
import org.fortiss.af3.platform.model.ExecutionUnit;
import org.fortiss.af3.platform.model.Route;
import org.fortiss.af3.platform.model.Segment;
import org.fortiss.af3.schedule.model.PeriodicTimeTrigger;
import org.fortiss.af3.schedule.model.ResourceAllocation;
import org.fortiss.af3.schedule.model.SystemSchedule;
import org.fortiss.af3.task.model.Signal;
import org.fortiss.af3.task.model.Task;
import org.fortiss.af3.task.model.TaskOutputPort;
import org.fortiss.af3.task.model.allocation.ComponentToTaskAllocationEntry;
import org.fortiss.af3.task.model.allocation.SignalToRouteAllocationEntry;
import org.fortiss.af3.task.model.allocation.TaskToExecutionUnitAllocationEntry;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.common.util.LambdaUtils;
import org.opt4j.core.Genotype;

import com.google.common.collect.BiMap;
import com.google.common.collect.HashBiMap;
import com.google.inject.Provides;
import com.microsoft.z3.Context;

/**
 * Generates a {@link StrictTTSchedule} that is constructed by calling the AF3 scheduler based on
 * Z3.
 * 
 * @author diewald
 */
// TODO(AD): Add the following parameters to the schedule: <IExecutionUnitAdapter,
// TaskExUMappingEntry, TaskExUMappingEncoding>
// This implies updating the registration with the feature.
public class SMTStrictTTDecoder extends DecoderModule<StrictTTSchedule> {

	/** Timeout for the SMT backend to find a solution. */
	private final static int smtTimeout = 1000; // in milliseconds

	/**
	 * Relative portion of the upper latency bound used as a stopping criterion for the SMT calls.
	 */
	private final static BigDecimal relBoundDistance = BigDecimal.valueOf(0.01);

	/**
	 * Transforms {@link TaskExUMappingEncoding} to a {@link SuperSet} of
	 * {@link TaskToExecutionUnitAllocationEntry}s.
	 */
	private TaskToExUTransformation allocTransformator = new TaskToExUTransformation();

	ScheduleRun smtSchedRun = new ScheduleRun();

	/** The SMT(Z3) scheduler. */
	// private ScheduleRun smtSchedRun = new ScheduleRun();

	/** See {@link SMTStrictTTDecoder}. */
	@Provides
	@Decodes
	public StrictTTSchedule decode(TaskExUMappingEncoding tmEnc, MessageEncoding msgEnc,
			PlatformCommunicationGraphEncoding pcgEnc, TaskMapper tMapper,
			ExplorationSpecification expSpec) {
		// TODO: Externalize the validation such that the encoding is not required to be passed to
		// the constructor (is there for validation purposes)
		BiMap<EObject, EObject> copyRefMap = HashBiMap.create();
		ExplorationSpecification copiedSpec = copyExplorationSpecification(expSpec, copyRefMap);
		SuperSetMap superSets = copiedSpec.getSearchSpace();

		SuperSet<TaskToExecutionUnitAllocationEntry> taskToExecUnitSet =
				allocTransformator.transform(tmEnc);
		superSets.put(TaskToExecutionUnitAllocationEntry.class, taskToExecUnitSet);

		MessageToRouteTransformation msgToRouteTransformer = new MessageToRouteTransformation();
		SuperSet<SignalToRouteAllocationEntry> sigToRoute = msgToRouteTransformer.transform(msgEnc);
		Collection<Route> routes =
				sigToRoute.getEntries().stream().map(e -> e.getRoute()).collect(toSet());
		superSets.put(Route.class, createSuperSet(routes, "Routes", Route.class));

		// TODO: Do not query the transmissionUnits from the input set but the platform graph.
		SuperSet<SignalToRouteAllocationEntry> sigToRouteOut =
				createSuperSet(emptyList(), "SignalToRoutes", SignalToRouteAllocationEntry.class);
		SystemSchedule sysSched = generateSchedule(superSets, sigToRouteOut, null);
		SuperSet<ResourceAllocation> rsAllocSet =
				createSuperSet("ResourceAllocations", ResourceAllocation.class);
		rsAllocSet.getEntries().addAll(getChildrenWithType(sysSched, ResourceAllocation.class));

		List<SMTConstraint> basicScheduleConstraints = generatePureScheduleConstraints(rsAllocSet,
				superSets.get(ComponentToTaskAllocationEntry.class), sigToRouteOut);

		// We have to add the scheduling constraints at the beginning of the target list.
		List<ExplorationTarget<?>> cachedTargets = copiedSpec.getTargets();
		copiedSpec.getTargets().clear();
		copiedSpec.getTargets().addAll(basicScheduleConstraints);

		SuperSetMap updSuperSets = copiedSpec.getSearchSpace();
		updSuperSets.put(ResourceAllocation.class, rsAllocSet);
		updSuperSets.put(SignalToRouteAllocationEntry.class, sigToRouteOut);

		ExplorationSolution expSol = seekApproxOptimalSchedule(copiedSpec, rsAllocSet);
		if(expSol == null) {
			StrictTTSchedule<IExecutionUnitAdapter, TaskExUMappingEntry, TaskExUMappingEncoding> schedule =
					new StrictTTSchedule<>(tmEnc);
			schedule.setInfeasible();
			return schedule;
		}

		return constructInternalSchedule(tmEnc, msgEnc, pcgEnc, tMapper, expSol);
	}

	/**
	 * Executes the SMT scheduler to find a solution with minimum latency. Therefore, a latency
	 * bound is encoded as a constraint and the result is checked for feasibility. In a loop, this
	 * bound is modified (1/2 step size) until the step size reached {@value #relBoundDistance} of
	 * the upper latency bound.
	 * 
	 * @param expSpec
	 *            {@link ExplorationSpecification} with {@link ResourceAllocation} skeleton.
	 * @param rsAllocSet
	 *            {@link ResourceAllocation} skeleton.
	 * @return Feasible {@link ExplorationSolution} with minimum latency.
	 */
	private ExplorationSolution seekApproxOptimalSchedule(ExplorationSpecification expSpec,
			SuperSet<ResourceAllocation> rsAllocSet) {
		ExplorationSolution latestValidSolution = null;
		ExplorationSolution expSol = null;

		BigDecimal upperBound = getUpperBoundScheduleLength(rsAllocSet);
		BigDecimal relUpperBound = upperBound.multiply(relBoundDistance);

		BigDecimal diffBound = upperBound;
		BigDecimal activeBound = upperBound;

		Map<String, String> parameters = new HashMap<>();
		parameters.put("timeout", smtTimeout + "");
		Context context = new Context(parameters);
		DSMLtoSMTTransformator transformation = new DSMLtoSMTTransformator();

		SolutionState solState = SolutionState.UNKNOWN;
		do {
			SMTConstraint constraint = createScheduleLengthConstraint(rsAllocSet, activeBound);
			expSpec.getTargets().add(constraint);
			expSol = createExplorationSolution();
			expSol.setExplorationSpec(expSpec);

			transformation.transform(expSpec, context);
			smtSchedRun.solveNonOptimized(context, parameters, transformation, expSol, null);

			expSpec.getTargets().remove(constraint);

			solState = expSol.getSolutionState();
			diffBound = diffBound.divide(BigDecimal.valueOf(2.0));
			switch(solState) {
				case UNSAT:
					activeBound = activeBound.add(diffBound);
					break;
				case OPTIMIZED:
				case OPTIMAL:
					activeBound = activeBound.subtract(diffBound);
					latestValidSolution = expSol;
					break;
				case UNKNOWN:
					// no solution within timeout --> drop the candidate
					return latestValidSolution;
			}
		} while(diffBound.doubleValue() > relUpperBound.doubleValue());

		return latestValidSolution;
	}

	/**
	 * Constructs the internal schedule representation from the returned {@link SuperSet} of
	 * {@link ResourceAllocation}s from the SMT solver.
	 */
	private StrictTTSchedule constructInternalSchedule(TaskExUMappingEncoding tmEnc,
			MessageEncoding msgEnc, PlatformCommunicationGraphEncoding pcgEnc, TaskMapper tMapper,
			ExplorationSolution schedSolution) {
		StrictTTSchedule<IExecutionUnitAdapter, TaskExUMappingEntry, TaskExUMappingEncoding> schedule =
				new StrictTTSchedule<>(tmEnc);

		// TODO(AD): Check solution state and document the "get(0)".
		SuperSet<ResourceAllocation> resAllocSol =
				schedSolution.getSolutions().get(0).getSolutionModel(ResourceAllocation.class);
		// TODO(AD): Check solution state and document the "get(0)".
		SuperSet<SignalToRouteAllocationEntry> sigAllocSol = schedSolution.getSolutions().get(0)
				.getSolutionModel(SignalToRouteAllocationEntry.class);

		// TODO(AD): Extract & centralize
		Map<IModelElement, IResourceAdapter> resourceMap = new HashMap<>();
		for(IResourceAdapter currResource : pcgEnc.getActualGraph().vertexSet()) {
			resourceMap.put(currResource.getObject(IModelElement.class), currResource);
		}

		Map<Segment, Message> segMsgMap = generateSegmenMsgMap(msgEnc, sigAllocSol);

		for(ResourceAllocation resAlloc : resAllocSol.getEntries()) {
			IModelElement schedME = resAlloc.getSchedulableEntity().getModelElement();
			// TODO(AD): Add handling of output ports / signals.
			IRequestAdapter requester = null;
			IResourceAdapter resource = null;
			if(schedME instanceof Task) {
				requester = tMapper.getAdapter((Task)schedME);
				// TODO(AD): Only 1:1 mappings are supported as of now.
				resource = getFirst(tmEnc.getResources((ITaskAdapter)requester)).get();
			} else if(schedME instanceof Segment) {
				IModelElement segmentResource = ((Segment)schedME).getEntity();
				resource = resourceMap.get(segmentResource);
				Assert.isNotNull(resource);
			}

			PeriodicTimeTrigger trigger = (PeriodicTimeTrigger)resAlloc.getTrigger();
			Double start = trigger.getStartTime().doubleValue();
			Double end = 5.0 + start.doubleValue();
			if(requester instanceof ITaskAdapter) {
				double duration =
						((ITaskAdapter)requester).getWcet((IExecutionUnitAdapter)resource);
				end = duration + start.doubleValue();

				// TODO: We only support 1:1 mappings in this place.
				TaskExUMappingEntry mappingEntry =
						getFirst(tmEnc.getMappingEntriesOf((ITaskAdapter)requester)).get();
				schedule.createTimeSlot(mappingEntry, resource, emptyList(), start, end, 0);
			} else {
				Message assocMessage = segMsgMap.get(schedME);
				isNotNull(assocMessage);
				TimeSlot communicationSlot = new TimeSlot(assocMessage,
						(ICommunicationResourceAdapter)resource, emptyList(), start, end);
				schedule.registerTimeSlot(communicationSlot, null, resource);
			}
		}

		return schedule;
	}

	/**
	 * Generates a mapping of {@link Segment}s to the {@link Message}s that were used to construct
	 * their parent {@link Route}s.
	 */
	private Map<Segment, Message> generateSegmenMsgMap(MessageEncoding msgEnc,
			SuperSet<SignalToRouteAllocationEntry> sigRouteAllocs) {
		Map<Segment, Message> segmentMsgMap = new HashMap<>();

		Map<IModelElement, Message> senderMsgMap = new HashMap<>();
		for(Message msg : msgEnc.getAllMessages()) {
			IModelElement senderPort = msg.getSignalAdapter().getObject(IModelElement.class);
			senderMsgMap.put(senderPort, msg);
		}

		for(SignalToRouteAllocationEntry sigRouteEntry : sigRouteAllocs.getEntries()) {
			Signal signal = LambdaUtils.getFirst(sigRouteEntry.getSignals()).get();
			TaskOutputPort sourceTaskPort = signal.getSourceTaskPort();
			Message msg = senderMsgMap.get(sourceTaskPort);

			Route route = sigRouteEntry.getRoute();
			Queue<Segment> segments = new LinkedList<>();
			segments.add(route.getRootSegment());
			while(!segments.isEmpty()) {
				Segment curSegment = segments.poll();
				segmentMsgMap.put(curSegment, msg);
				segments.addAll(curSegment.getNext());
			}
		}

		return segmentMsgMap;
	}

	/**
	 * Returns the upper bound of a time-triggered schedule for the given {@link ResourceAllocation}
	 * {@link SuperSet} (= sum of all {@link Task} WCETs).
	 */
	private static BigDecimal getUpperBoundScheduleLength(SuperSet<ResourceAllocation> rsAllocSet) {
		BigDecimal upperBound = BigDecimal.ZERO;
		for(ResourceAllocation ra : rsAllocSet.getEntries()) {
			IModelElement modelElement = ra.getSchedulableEntity().getModelElement();
			if(!(modelElement instanceof Task)) {
				continue;
			}
			Task task = (Task)modelElement;
			BigDecimal wcet = task.getWcet((ExecutionUnit)ra.getResource());
			upperBound = upperBound.add(wcet);
		}
		return upperBound;
	}

	/**
	 * Creates an {@link SMTConstraint} that enforces that the end time of all
	 * {@link ResourceAllocation}s is below a given {@code upperBound}.
	 */
	private static SMTConstraint createScheduleLengthConstraint(
			SuperSet<ResourceAllocation> resAllocSS, BigDecimal upperBound) {
		org.fortiss.af3.exploration.dseml.model.expression.Set<ResourceAllocation> set =
				createSet(resAllocSS, "AllResources_" + resAllocSS.getName() + "_Set",
						ResourceAllocation.class);
		set.getEntries().addAll(resAllocSS.getEntries());

		ModelElementPropertyLiteral startSet = createStartTimeLiteral(set);
		ModelElementPropertyLiteral durationSet = createDurationLiteral(set);
		Plus endTime = createPlus(startSet, durationSet);

		LessEqual lessEqual = createLessEqual(endTime, createArithmeticLiteral(upperBound));
		ForAll forAll = generateForAll(set, lessEqual);

		return createSMTConstraint("Latency: " + upperBound, forAll, true);
	}

	/**
	 * Creates an objective that minimizes the sum of ending times of tasks, for the given
	 * {@link SystemSchedule}.
	 */
	// TODO(AD): Extend for the per-resource case.
	public static SMTObjective
			createMinSumEndTimesObjective(SuperSet<ResourceAllocation> resAllocSS) {
		Set<ResourceAllocation> set =
				createSet(resAllocSS, "aResourceAllocationsSet", ResourceAllocation.class);
		set.getEntries().addAll(resAllocSS.getEntries());

		ModelElementPropertyLiteral startLit = createStartTimeLiteral(set);
		ModelElementPropertyLiteral durationLit = createDurationLiteral(set);
		Plus endTime = createPlus(startLit, durationLit);
		BooleanLiteral trivialCondition = createBooleanLiteral(true);
		Sum sum = createSum(set.getCastedSet(IModelElement.class), trivialCondition, endTime);

		Minimize min = createMinimize(sum);
		return createSMTObjective("Minimize Latency", min);
	}

	/** {@inheritDoc} */
	@Override
	public StrictTTSchedule<IExecutionUnitAdapter, TaskExUMappingEntry, TaskExUMappingEncoding>
			decode(Genotype genotype) {
		// Required by the framework...
		return null;
	}
}
