/*-------------------------------------------------------------------------+
| Copyright 2015 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.feature.scheduling.evaluator;

import static java.lang.Math.pow;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.evaluate.annotations.Evaluate;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.evaluate.annotations.Validates;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.ITaskGraph;
import org.fortiss.af3.exploration.alg.feature.scheduling.StrictTTSchedule;
import org.fortiss.af3.exploration.alg.feature.taskmapping.problem.TaskGraphEncoding;
import org.fortiss.af3.exploration.alg.feature.taskmapping.variable.TaskExUMappingEncoding;
import org.fortiss.af3.exploration.alg.feature.taskmapping.variable.TaskMappingEncodingBase;
import org.fortiss.af3.exploration.alg.feature.taskmapping.variable.TaskMappingEntry;
import org.fortiss.af3.exploration.alg.module.EvaluatorModule;
import org.fortiss.af3.exploration.model.ExplorationConstraint;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.fortiss.af3.exploration.moea.model.constraints.PeriodConstraint;
import org.opt4j.core.DoubleValue;
import org.opt4j.core.Value;

import com.google.inject.Provides;

/**
 * This constraint evaluator evaluates whether all {@link ITaskAdapter}s are
 * executed within their respective period. Any violation is summed and then squared.
 * 
 * @author diewald
 */
public class PeriodConstraintEvaluator<T extends IExecutionUnitAdapter, E extends TaskMappingEntry<ITaskAdapter, T>, M extends TaskMappingEncodingBase<ITaskAdapter, T, E, ?>>
		extends EvaluatorModule<Double, PeriodConstraint, StrictTTSchedule<T, E, M>> {

	/** Objective to track evaluation of this {@link ExplorationTarget}. */
	public PeriodConstraintEvaluator(Class<PeriodConstraint> optGoal) {
		super(optGoal);
	}

	/** Objective to track evaluation of this {@link ExplorationTarget}. */
	public PeriodConstraintEvaluator(PeriodConstraint optGoal) {
		super(optGoal);
	}

	/**
	 * evaluate the deadline objective
	 * if the deadline is met, we set the fitness value to -1
	 * if the deadline is violated, we set the fitness value to the gap between the actual latency
	 * and the deadline
	 */
	@Provides
	@Evaluate(PeriodConstraintEvaluator.class)
	public DoubleValue evaluate(@SuppressWarnings("rawtypes") StrictTTSchedule schedule,
			TaskGraphEncoding tgEnc, TaskExUMappingEncoding itmEnc) {
		double retval = 0;

		// we evaluate if all tasks are within the period. we want everything to be
		// finished within the period
		for(ITaskAdapter deployableComponent : itmEnc.getRequesters()) {
			ITaskGraph<ITaskAdapter> taskGraph = tgEnc.getTaskGraphOf(deployableComponent);
			// we need to check each iteration in the hyper period
			if(taskGraph != null) {
				for(int i = 0; i < tgEnc.getHyperPeriod() / taskGraph.getHyperPeriod(); i++) {
					double periodEnd = (i + 1) * taskGraph.getHyperPeriod();
					double end = schedule.getLastEndTime(deployableComponent, i);
					if(end > periodEnd) {
						retval += end - periodEnd;
					}
				}
			}
		}

		return new DoubleValue(retval > 0 ? pow(retval, 2) : 0);
	}

	/** Validates the {@link ExplorationConstraint} versus the penalty value. */
	@Provides
	@Validates
	public Boolean validate(Value<?> value) {
		Double evalValue = value.getDouble();
		Double limit = explorationTarget.getPeriod();
		if(evalValue > limit + 10e-6) {
			return false;
		}
		return true;
	}
}
