/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.feature.scheduling.scheduler;

import static java.util.Arrays.asList;
import static org.eclipse.core.runtime.Assert.isTrue;
import static org.fortiss.af3.exploration.alg.util.ExplorationAlgUtils.DSE_VALIDATE_SCHEDULE;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.TreeSet;

import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ICommunicationResourceAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IGatewayUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITransmissionUnitAdapter;
import org.fortiss.af3.exploration.alg.feature.platform.PlatformCommunicationGraphEncoding;
import org.fortiss.af3.exploration.alg.feature.scheduling.ResourceSchedule;
import org.fortiss.af3.exploration.alg.feature.scheduling.StrictTTSchedule;
import org.fortiss.af3.exploration.alg.feature.scheduling.TimeSlot;
import org.fortiss.af3.exploration.alg.feature.taskmapping.problem.comm.IMessageRouter;
import org.fortiss.af3.exploration.alg.feature.taskmapping.problem.comm.Message;
import org.fortiss.af3.exploration.alg.feature.taskmapping.problem.comm.ShortestPathRouter;
import org.fortiss.af3.exploration.alg.feature.taskmapping.variable.TaskMappingEntry;
import org.jgrapht.graph.AbstractGraph;
import org.jgrapht.graph.DefaultEdge;

import com.google.common.collect.HashMultimap;
import com.google.common.collect.Multimap;

/**
 * Simple TDMA bus scheduler that adds the communication time to a pure task schedule
 * ({@link StrictTTSchedule}).
 * 
 * @author huang
 */
public class TDMABusScheduler extends SchedulerBase {

	/**
	 * References the {@link IMessageRouter} used to determine the path of generated {@link Message}
	 * s within this scheduler.
	 */
	private IMessageRouter messageRouter;

	/** Constructs a {@link TDMABusScheduler} for a given input model. */
	public TDMABusScheduler() {
		this.messageRouter = new ShortestPathRouter();
	}

	/** Add simple TDMA bus schedule / communication time to a pure task schedule. */
	public synchronized void schedule(StrictTTSchedule<?, ?, ?> schedule,
			Collection<Message> generatedMessages, PlatformCommunicationGraphEncoding pcgEncoding) {
		Multimap<TimeSlot, Message> emitableMessagesPerSlot = HashMultimap.create();
		Map<TimeSlot, Integer> reqRecvMsgsPerSlot = new HashMap<TimeSlot, Integer>();
		// Add the information about the allocated TimeSlots for the senders and the receivers of
		// the messages to the generated messages. This information stems from the schedule of the
		// the execution units
		Map<Message, AbstractGraph<IResourceAdapter, DefaultEdge>> messageRoutes =
				new HashMap<Message, AbstractGraph<IResourceAdapter, DefaultEdge>>();
		for(Message message : generatedMessages) {
			TimeSlot senderSlot = schedule.getTimeSlotOf(message.getSender());
			if(senderSlot == null) {
				// Message is not schedulable (e.g. an RC message) --> Skip scheduling this message.
				continue;
			}
			message.addMessageSenderTimeSlot(senderSlot);
			emitableMessagesPerSlot.put(senderSlot, message);
			for(TaskMappingEntry<?, ?> receiverEntry : message.getReceiverMappingEntries()) {
				TimeSlot recvTimeSlot = schedule.getTimeSlotOf(receiverEntry);
				message.addMessageReceiverTimeSlots(recvTimeSlot);
				Integer numReqMsgs = reqRecvMsgsPerSlot.get(recvTimeSlot);
				if(numReqMsgs == null) {
					numReqMsgs = Integer.valueOf(0);
				}
				numReqMsgs++;
				reqRecvMsgsPerSlot.put(recvTimeSlot, numReqMsgs);
			}
		}

		// Construct a message queue whose ordering defines the type of scheduling, i.e. ESF, EDF
		// etc. It is initially filled with the generated messages resulting from the encoding and
		// also captures messages generated by this message scheduler.
		TreeSet<Message> messageReleaseQueue = new TreeSet<>();
		TreeSet<Message> messageReadyQueue = new TreeSet<>();
		for(Message readyMessage : generatedMessages) {
			TimeSlot senderSlot = readyMessage.getSenderTimeSlot();
			if(senderSlot == null) {
				// Message is not schedulable (e.g. an RC message) --> Skip scheduling this message.
				continue;
			}

			if(!anyPredecessorSlotRequiresMsgs(senderSlot, reqRecvMsgsPerSlot) &&
					reqRecvMsgsPerSlot.get(senderSlot) == null) {
				messageReadyQueue.add(readyMessage);
			} else {
				messageReleaseQueue.add(readyMessage);
			}
		}

		// Process all messages present in the schedule. Note that new messages are generated during
		// message scheduling as a consequence of multi-hop messaging.
		while(!messageReadyQueue.isEmpty()) {
			// Update the currently considered message with the one that has the earliest start
			// time.
			Message currentMessage = messageReadyQueue.pollFirst();

			// First, we need to retrieve the next communication resource along which the current
			// message being routed. This is the second element in the list of participating
			// communication resources, since the first one is the sender.
			AbstractGraph<IResourceAdapter, DefaultEdge> messageGraph = messageRouter
					.calculateMessageRoute(currentMessage.getSenderTimeSlot().getResource(),
							currentMessage.getReceiverResources(), pcgEncoding);
			// messageRoutes.get(currentMessage);

			Set<DefaultEdge> senderResourceEdges =
					messageGraph.outgoingEdgesOf(currentMessage.getSenderTimeSlot().getResource());
			// Here, we assume that a message may have only one source element. This assumption fits
			// to multi-cast messaging and the principle how new messages are generated.
			isTrue(senderResourceEdges.size() == 1,
					"Each message must have exactly one source slot!");
			if(messageGraph.getEdgeTarget(
					senderResourceEdges.iterator().next()) instanceof IExecutionUnitAdapter) {
				pcgEncoding.display();
			}
			ICommunicationResourceAdapter communicationResource =
					(ICommunicationResourceAdapter)messageGraph
							.getEdgeTarget(senderResourceEdges.iterator().next());

			// Create a time slot for the current message in the schedule of the next
			// communication resource (of the route).
			TimeSlot communicationSlot = createCommunicationTimeSlot(schedule, currentMessage,
					currentMessage.getSenderTimeSlot(), communicationResource);

			// NOTE: Here, the sender slot can also be the slot of a previous message.
			TimeSlot senderSlot = currentMessage.getSenderTimeSlot();
			for(TimeSlot successorSlot : currentMessage.getReceiverTimeSlots()) {
				communicationSlot.addSuccessor(successorSlot);
				senderSlot.replaceSuccessor(successorSlot, communicationSlot);
				successorSlot.replacePredecessor(senderSlot, communicationSlot);
			}
			// It is required to shift the receiver tasks by the amount of time
			// that the communicationSlot requires for transmission. This is needed to
			// maintain causality.
			shiftReceiverSlot(schedule, currentMessage, communicationSlot);
			// Determine and generate successor messages from the current one.

			Collection<TimeSlot> reachedReceiverSlots = new HashSet<TimeSlot>();
			Collection<Message> succMessages = generateSuccessorMessages(schedule, currentMessage,
					messageGraph, communicationResource, communicationSlot, messageRoutes,
					reqRecvMsgsPerSlot, reachedReceiverSlots);
			messageReadyQueue.addAll(succMessages);
			// The ready and release queues are updated for the next iteration.
			List<TimeSlot> readySenderSlots = findNextSenderSlots(reachedReceiverSlots,
					reqRecvMsgsPerSlot, emitableMessagesPerSlot);
			updateMessageQueues(readySenderSlots, emitableMessagesPerSlot, messageReleaseQueue,
					messageReadyQueue);
		}
		if(!messageReleaseQueue.isEmpty()) {
			schedule = null;
		}
	}

	/**
	 * Determines whether the given {@link TimeSlot} or any of its predecessors must wait for
	 * {@link Message}s to arrive such that they can be executed. Therefore a map is required which
	 * counts the number of remaining {@link Message}s required as input for each {@link TimeSlot}.
	 * Note that {@link TimeSlot}s that do not require {@link Message}s as input may not be present
	 * in the map.
	 * <p>
	 * Implementation note: This method considers the case at which all predecessors are located on
	 * the same {@link IExecutionUnitAdapter} as the given {@link TimeSlot}, i.e. at which no
	 * {@link Message}s are required to pass the output from the predecessors to the input of the
	 * given successor {@link TimeSlot}.
	 * 
	 * @param msgSenderSlot
	 *            {@link TimeSlot} whose predecessors shall be checked for requiring {@link Message}
	 *            s (typically the sender {@link TimeSlot} of a {@link Message}.
	 * @param reqRecvMsgsPerSlot
	 *            Map containing information about the remaining {@link Message}s that are required
	 *            to execute a specific {@link TimeSlot} (-> key).
	 * @return Boolean indicating if any (relevant) predecessor is waiting for a {@link Message} to
	 *         be executed.
	 */
	private boolean anyPredecessorSlotRequiresMsgs(TimeSlot msgSenderSlot,
			Map<TimeSlot, Integer> reqRecvMsgsPerSlot) {
		LinkedList<TimeSlot> predecessorList = new LinkedList<TimeSlot>();
		predecessorList.addAll(msgSenderSlot.getPredecessors());
		while(!predecessorList.isEmpty()) {
			TimeSlot currentPredescessor = predecessorList.poll();
			if(currentPredescessor.getResource() != msgSenderSlot.getResource()) {
				predecessorList.addAll(currentPredescessor.getPredecessors());
				if(reqRecvMsgsPerSlot.get(currentPredescessor) != null &&
						reqRecvMsgsPerSlot.get(currentPredescessor) > 0) {
					return true;
				}
			}
		}

		return false;
	}

	// TODO: another notion for messages is needed: Eliminate the message route map (annotate
	// directly)
	/**
	 * Generates (and registers) new messages for the remaining path of the message. If there are
	 * more than one resource "following" the just considered resource, multiple messages are
	 * generated.
	 * 
	 * @param schedule
	 *            The schedule for which messages shall be generated.
	 * @param currentMessage
	 *            The message for which its successors shall be generated.
	 * @param currentMessageGraph
	 *            The graph containing the path of the message in the platform.
	 * @param communicationResource
	 *            The communication {@link IResourceAdapter} for which the successor messages shall
	 *            be generated. This is typically the (single) successor vertex of the source of the
	 *            current message graph.
	 * @param communicationSlot
	 *            The schedule time slot which identifies the sender of the generated messages.
	 * @param messageRoutes
	 *            Map that associates {@link Message}s with the (remaining) routes/paths they need
	 *            to traverse.
	 * @param reqRecvMsgsPerSlot
	 *            Stores the number of messages a {@link TimeSlot} (of an
	 *            {@link IExecutionUnitAdapter} must receive in order to be executed. Used for
	 *            slot-shifting.
	 * @param reachedReceiverSlots
	 *            Collection of {@link TimeSlot}s that received a message.
	 * @return Collection of {@link Message}s that must be sent from the currently traversed
	 *         Resource such that the signal arrives at the target.
	 */
	private Collection<Message> generateSuccessorMessages(StrictTTSchedule<?, ?, ?> schedule,
			Message currentMessage,
			AbstractGraph<IResourceAdapter, DefaultEdge> currentMessageGraph,
			ICommunicationResourceAdapter communicationResource, TimeSlot communicationSlot,
			Map<Message, AbstractGraph<IResourceAdapter, DefaultEdge>> messageRoutes,
			Map<TimeSlot, Integer> reqRecvMsgsPerSlot, Collection<TimeSlot> reachedReceiverSlots) {
		Collection<Message> generatedMessages = new HashSet<Message>();

		// From the receiver time slots representing the logical receivers of the current message,
		// create a collection of the (platform) resources that onto which these time slots are
		// deployed.
		Collection<TimeSlot> currentMessageRecieverSlots = currentMessage.getReceiverTimeSlots();
		Collection<IResourceAdapter> currentMessageReceiverResources =
				new ArrayList<IResourceAdapter>();
		for(TimeSlot currentReceiverSlot : currentMessageRecieverSlots) {
			currentMessageReceiverResources.add(currentReceiverSlot.getResource());
		}

		// Retrieve the routes of the successor messages
		Map<IResourceAdapter, AbstractGraph<IResourceAdapter, DefaultEdge>> remainingRoutes =
				messageRouter.getRemainingMessageGraphs(currentMessage, communicationResource,
						currentMessageGraph);
		// Generate a new message for each direct successor in the communication graph.
		for(Entry<IResourceAdapter, AbstractGraph<IResourceAdapter, DefaultEdge>> remRoute : remainingRoutes
				.entrySet()) {
			IResourceAdapter nextResource = remRoute.getKey();
			// If the next resource isn't a communication unit, we reached a resource and
			if(!(nextResource instanceof IExecutionUnitAdapter) &&
					!currentMessageReceiverResources.contains(nextResource)) {
				// Identify the route route corresponding to the current successor and generate a
				// list of receiver resources.
				AbstractGraph<IResourceAdapter, DefaultEdge> successorRoute =
						remainingRoutes.get(nextResource);
				HashSet<TimeSlot> successorReceiverSlots = new HashSet<TimeSlot>();
				for(TimeSlot receiverSlot : currentMessageRecieverSlots) {
					if(successorRoute.containsVertex(receiverSlot.getResource())) {
						successorReceiverSlots.add(receiverSlot);
						if(receiverSlot.getResource() instanceof IExecutionUnitAdapter) {
							System.out.println("Here");
						}
					}
				}

				// Finally, generate the message.
				if(!successorReceiverSlots.isEmpty()) {
					Message successorMessage = new Message(currentMessage.getSignalAdapter(),
							communicationSlot, successorReceiverSlots);
					generatedMessages.add(successorMessage);
					messageRoutes.put(successorMessage, successorRoute);
				}
			} else {
				// Book keeping for the messages that have received the target TimeSlots.
				for(TimeSlot currentReceiverSlot : currentMessageRecieverSlots) {
					if(currentReceiverSlot.getResource() == nextResource) {
						reachedReceiverSlots.add(currentReceiverSlot);
						Integer remainingMsgCount = reqRecvMsgsPerSlot.get(currentReceiverSlot);
						if(remainingMsgCount != null) {
							remainingMsgCount--;
							reqRecvMsgsPerSlot.put(currentReceiverSlot, remainingMsgCount);
						}
					}
				}
			}
		}

		return generatedMessages;
	}

	/**
	 * From a given collection of {@link TimeSlot}s that have received a {@link Message}, this
	 * method determines those {@link TimeSlot}s that are ready to emit {@link Message}s themselves.
	 * In order to determine these {@link TimeSlot}s a map that counts the missing number of
	 * {@link Message}s that must be received per {@link TimeSlot} and a map referencing the
	 * {@link Message}s that are emitted per {@link TimeSlot} are required.
	 * 
	 * @param reachedReceiverSlots
	 *            {@link TimeSlot}s that have received a {@link Message} in the current iteration of
	 *            the schedule.
	 * @param reqRecvMsgsPerSlot
	 *            Map counting the remaining number of required messages to execute the receiver
	 *            {@link TimeSlot}.
	 * @param emitableMessagesPerSlot
	 *            Map referencing the {@link Message} which are emitted after executing the
	 *            associated {@link TimeSlot} (-> key).
	 * @return List of {@link TimeSlot}s that may emit {@link Message}.
	 */
	private List<TimeSlot> findNextSenderSlots(Collection<TimeSlot> reachedReceiverSlots,
			Map<TimeSlot, Integer> reqRecvMsgsPerSlot,
			Multimap<TimeSlot, Message> emitableMessagesPerSlot) {
		List<TimeSlot> nextReadySenderSlots = new ArrayList<TimeSlot>();

		for(TimeSlot receiverSlot : reachedReceiverSlots) {
			LinkedList<TimeSlot> possiblyReadySuccessors = new LinkedList<TimeSlot>();
			possiblyReadySuccessors.add(receiverSlot);

			while(!possiblyReadySuccessors.isEmpty()) {
				TimeSlot currentSuccessor = possiblyReadySuccessors.poll();
				Integer requiredMsgNumForCurrentSuccessor =
						reqRecvMsgsPerSlot.get(currentSuccessor);
				if(requiredMsgNumForCurrentSuccessor != null &&
						requiredMsgNumForCurrentSuccessor == 0) {
					nextReadySenderSlots.add(currentSuccessor);
				} else if(requiredMsgNumForCurrentSuccessor == null) {
					possiblyReadySuccessors.addAll(currentSuccessor.getSuccessors());
				}
			}
		}

		return nextReadySenderSlots;
	}

	/**
	 * Modifies the given ready and release queues such that the (now) emittable messages are moved
	 * to the ready queue and are eliminated from the release queue.
	 * 
	 * @param readySenderSlots
	 *            List of {@link TimeSlot}s which are ready to emit {@link Message}s, i.e.
	 *            {@link TimeSlot}s which themselves (or their predecessors) have received all
	 *            required {@link Message}s those {@link TimeSlot}s require as input.
	 * @param emittableMessagesPerSlot
	 *            Map containing the {@link Message}s each {@link TimeSlot} emits, once it is ready.
	 * @param messageReleaseQueue
	 *            Queue containing the non-ready {@link Message}s.
	 * @param messageQueueReady
	 *            Queue containing the {@link Message}s that can be scheduled.
	 */
	private void updateMessageQueues(List<TimeSlot> readySenderSlots,
			Multimap<TimeSlot, Message> emittableMessagesPerSlot,
			TreeSet<Message> messageReleaseQueue, TreeSet<Message> messageQueueReady) {

		for(TimeSlot currentSlot : readySenderSlots) {
			Collection<Message> readyMessages = emittableMessagesPerSlot.get(currentSlot);
			messageQueueReady.addAll(readyMessages);
			messageReleaseQueue.removeAll(readyMessages);
		}
	}

	/**
	 * Creates the {@link TimeSlot} on a communication resource (like an
	 * {@link ITransmissionUnitAdapter})
	 * for a given Message.
	 * 
	 * @param schedule
	 *            The currently considered schedule.
	 * @param message
	 *            Message, for which the slot has to be created.
	 * @param predecessorSlot
	 *            predecessor {@link TimeSlot} (sender or preceding {@link Message} segment).
	 * @param communicationResource
	 *            An {@link ITransmissionUnitAdapter} or {@link IGatewayUnitAdapter} in whose
	 *            schedule the {@link TimeSlot} is about being registered.
	 * @return The {@link TimeSlot} that has just been added to the schedule.
	 */
	private TimeSlot createCommunicationTimeSlot(StrictTTSchedule<?, ?, ?> schedule,
			Message message, TimeSlot predecessorSlot,
			ICommunicationResourceAdapter communicationResource) {
		// final double minTransmissionDuration = 0.1;
		// Constant used for bit->MByte transformation.
		final double bitsPerMByte = 8 * 1024 * 1024;
		double messageStartTime = message.getSenderTimeSlot().getEndTime();

		// Determine the time required for the transmission of the message. A minimal transmission
		// time is added to avoid time slots with a duration of zero.
		double resourceWCTT = 0.0;
		if(communicationResource instanceof ITransmissionUnitAdapter) {
			resourceWCTT = ((ITransmissionUnitAdapter)communicationResource)
					.getTransmissionDuration(message.getMessageSize() / bitsPerMByte);
		} else if(communicationResource instanceof IGatewayUnitAdapter) {
			resourceWCTT = ((IGatewayUnitAdapter)communicationResource).getWCFT();
		}
		resourceWCTT = Math.max(resourceWCTT, 1.0);

		// Create a new slot for the message.
		TimeSlot communicationSlot = new TimeSlot(message, communicationResource,
				asList(predecessorSlot), messageStartTime, messageStartTime + resourceWCTT);

		// Insert the created slot into the schedule of the communication resource.
		schedule.registerTimeSlot(communicationSlot, null, communicationResource);
		shiftSlotToEarliestPossibleTime(schedule.getResourceSchedule(communicationResource),
				communicationSlot);

		return communicationSlot;
	}

	/**
	 * Moves the given {@link TimeSlot} to the earliest 'position' in the given
	 * {@link ResourceSchedule} where it does not overlap with other {@link TimeSlot}s of the target
	 * {@link IResourceAdapter}.
	 * 
	 * @param resourceSchedule
	 *            The {@link ResourceSchedule} containing the {#link insertSlot}.
	 * @param insertSlot
	 *            The {@link TimeSlot} that shall be shifted to a non-overlapping 'position' of the
	 *            given {@link ResourceSchedule}.
	 */
	private void shiftSlotToEarliestPossibleTime(ResourceSchedule resourceSchedule,
			TimeSlot insertSlot) {
		double earliestPossibleStartTime = insertSlot.getStartTime();

		// Determine the slots of the resource schedule that are temporally located 'behind' the
		// start time of the slot to be inserted since these slots may impact the 'position' of the
		// slot to be inserted.
		TreeSet<TimeSlot> orderedResourceSchedule = new TreeSet<TimeSlot>();
		orderedResourceSchedule.addAll(resourceSchedule.keySet());
		TreeSet<TimeSlot> relevantTimeSlots = new TreeSet<TimeSlot>();
		relevantTimeSlots.addAll(orderedResourceSchedule.tailSet(insertSlot));

		// Depending on the existence of a temporally preceding (by the start time) slot, determine
		// the earliest possible start time of the slot to be inserted.
		// - lower(...) returning null implies no preceding slot present
		// - otherwise: the earliest possible start time is that of the inserted slot or the 'end
		// time' of the preceding slot
		if(orderedResourceSchedule.lower(insertSlot) != null) {
			earliestPossibleStartTime =
					Math.max(orderedResourceSchedule.lower(insertSlot).getEndTime(),
							insertSlot.getStartTime());
		}

		// Find the earliest possible start time for the slot to be inserted that is determined by a
		// sufficiently large 'time gap' relative to already present slots in the schedule.
		// NOTE: More "beautiful" solutions are possible, but cause errors if double values are used
		// due to precision errors during variable assignment.
		double timeGap = 0.0;
		for(TimeSlot nextTimeSlot : relevantTimeSlots) {
			timeGap = nextTimeSlot.getStartTime() - earliestPossibleStartTime;
			if(timeGap > insertSlot.getDuration()) {
				break;
			}
			earliestPossibleStartTime = nextTimeSlot.getEndTime();
		}

		// Perform the shift of the slot to the earliest valid start time.
		double shiftTime = earliestPossibleStartTime - insertSlot.getStartTime();
		resourceSchedule.shiftSlotAndSuccessors(insertSlot, shiftTime);
	}

	/**
	 * Shifts {@link TimeSlot}s that are receivers of the given message backwards on their
	 * respective {@link ResourceSchedule}, if required. The amount of time by which the receiving
	 * {@link TimeSlot}s are shifted is determined by the 'end time' of the {#link
	 * communicationSlot}.
	 * Subsequent slots, that are affected by the shift of a receiver, are also shifted until there
	 * exists no more temporal overlap between time slots.
	 * 
	 * @param schedule
	 *            The currently considered schedule.
	 * @param shiftMessage
	 *            The message
	 * @param communicationSlot
	 *            The {@link TimeSlot} allocated at a {@link ICommunicationResourceAdapter} whose
	 *            insertion (possibly) triggers the shift of other {@link TimeSlot}s.
	 */
	private void shiftReceiverSlot(StrictTTSchedule<?, ?, ?> schedule, Message shiftMessage,
			TimeSlot communicationSlot) {
		List<TimeSlot> shiftedSlotsBookkeeping = new ArrayList<>();

		// List of successor TimeSlots that must be considered in the successive shifts.
		LinkedList<TimeSlot> succTimeSlots = new LinkedList<TimeSlot>();
		succTimeSlots.addAll(communicationSlot.getSuccessors());

		while(!succTimeSlots.isEmpty()) {
			TimeSlot recvSlot = succTimeSlots.poll();
			Optional<TimeSlot> latestPredSlot = recvSlot.getPredecessors().stream()
					.max((s1, s2) -> Double.compare(s1.getEndTime(), s2.getEndTime()));
			double latestPredEndTime =
					latestPredSlot.isPresent() ? latestPredSlot.get().getEndTime() : 0.0;
			double minShift = latestPredEndTime - recvSlot.getStartTime();
			if(minShift > 0.0) {
				shiftedSlotsBookkeeping.add(recvSlot);
				ResourceSchedule resSched = schedule.getResourceSchedule(recvSlot.getResource());
				// TODO(AD): HACK. Check operators.
				if(resSched == null) {
					return;
				}
				Collection<TimeSlot> shiftedSlots =
						resSched.shiftSlotAndSuccessors(recvSlot, minShift);
				succTimeSlots.addAll(shiftedSlots);
				shiftedSlots.forEach(s -> succTimeSlots.addAll(s.getSuccessors()));
			}
		}

		if(DSE_VALIDATE_SCHEDULE) {
			isTrue(schedule.validateSchedule(true));
		}
	}
}
