/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.feature.taskmapping.problem.comm;

import java.util.Collection;
import java.util.Map;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.decode.Decodes;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.exception.DecodingException;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.feature.partition.variable.TaskPartitionMappingEncoding;
import org.fortiss.af3.exploration.alg.feature.platform.PlatformCommunicationGraphEncoding;
import org.fortiss.af3.exploration.alg.feature.taskinstantiation.problem.InstantiatedTaskGraphEncoding;
import org.fortiss.af3.exploration.alg.feature.taskmapping.problem.TaskGraphEncoding;
import org.fortiss.af3.exploration.alg.feature.taskmapping.problem.comm.IMessageRouter.RoutingStrategy;
import org.fortiss.af3.exploration.alg.feature.taskmapping.variable.TaskMappingEncodingBase;
import org.fortiss.af3.exploration.alg.feature.taskmapping.variable.TaskMappingEntry;
import org.fortiss.af3.exploration.alg.module.DecoderModule;
import org.fortiss.af3.exploration.alg.module.common.encoding.IMappingEncoding;
import org.jgrapht.graph.DefaultDirectedGraph;
import org.jgrapht.graph.DefaultEdge;
import org.opt4j.core.Genotype;
import org.opt4j.core.problem.Decoder;

import com.google.inject.Provides;

/**
 * {@link Decoder} that takes an {@link TaskMappingEncodingBase} and a) an
 * generates the required messages for this {@link TaskMappingEncodingBase} and b) routes
 * the generated {@link Message} on the target platform.
 * 
 * @author diewald
 */
public class MessageDecoder<T extends IExecutionUnitAdapter, E extends TaskMappingEntry<ITaskAdapter, T>, M extends TaskMappingEncodingBase<ITaskAdapter, T, E, ?>>
		extends DecoderModule<MessageEncoding> {
	/**
	 * {@link MessageGenerator} which analyzes a {@link TaskMappingEncodingBase} and
	 * determines the required {@link Message}s.
	 */
	protected MessageGenerator<T, E, M> messageGenerator;

	/** References the {@link IMessageRouter} selected in the Constructor. */
	protected IMessageRouter messageRouter;

	/** Default constructor that requires a reference to the input models. */
	public MessageDecoder() {
		messageGenerator = new MessageGenerator<T, E, M>();
		messageRouter = new MessageRouter();
	}

	/**
	 * Produces a {@link MessageEncoding} that contains the multicast messages that are required to
	 * exchange signals between {@link IExecutionUnitAdapter}s. For each signal that is exchanged
	 * between two tasks or more (only one sender) that are located on different
	 * {@link IExecutionUnitAdapter}s, a {@link Message} is generated.
	 * 
	 * @param tmEnc
	 *            {@link TaskPartitionMappingEncoding} defining the "location" of
	 *            {@link ITaskAdapter}s.
	 * @param pcgEnc
	 *            Graph of the platform defining potential routes for {@link Message}s.
	 * @param itgEnc
	 *            {@link InstantiatedTaskGraphEncoding} containing the signals exchanged between
	 *            {@link ITaskAdapter}s.
	 * @return {@link MessageEncoding} that contains the multicast messages that are required to
	 *         exchange signals between {@link IExecutionUnitAdapter}s
	 * @throws DecodingException
	 */
	@SuppressWarnings("unchecked")
	@Provides
	@Decodes
	public MessageEncoding decode(
			IMappingEncoding<ITaskAdapter, IExecutionUnitAdapter, TaskMappingEntry<ITaskAdapter, IExecutionUnitAdapter>, ?> tmEnc,
			PlatformCommunicationGraphEncoding pcgEnc, TaskGraphEncoding itgEnc)
			throws DecodingException {
		Collection<Message> generatedMessages = messageGenerator.generateMessages((M)tmEnc, itgEnc);
		Map<Message, DefaultDirectedGraph<IResourceAdapter, DefaultEdge>> messageRouteMap =
				messageRouter.calculateMessageRoutes(generatedMessages, pcgEnc,
						RoutingStrategy.SHORTEST_PATH);
		if(messageRouteMap == null) {
			// At least one *invalid* communication link is required by the given mapping =>
			// infeasible mapping.
			throw new DecodingException(tmEnc, this,
					"At least one message could not be routed from the sending unit to all its receivers.");
		}

		// Add the calculated routes to the generated Messages.
		for(Message msg : generatedMessages) {
			msg.setRoute(messageRouteMap.get(msg));
		}

		return new MessageEncoding(generatedMessages);
	}

	/** {@inheritDoc} */
	@Override
	public MessageEncoding decode(Genotype genotype) {
		return null;
	}
}
