/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.service.module;

import java.util.function.Predicate;

import org.fortiss.af3.exploration.alg.exception.ExplorationException;
import org.fortiss.af3.exploration.alg.service.IExpSpecTransformationService;
import org.fortiss.af3.exploration.alg.service.IExplorationExtension;
import org.fortiss.af3.exploration.model.ExplorationSpecification;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.fortiss.af3.exploration.moea.model.constraints.InternalConstraint;

import com.google.inject.Injector;

/**
 * Module that defines transformations of the given input {@link ExplorationSpecification} that are
 * executed before the D3SE is launched. They have to be registered with the
 * {@link IExpSpecTransformationService} or other {@link IExpSpecTransformationModule}s
 * (composition) in order to get executed.
 * 
 * @author diewald
 */
public interface IExpSpecTransformationModule<I, O> extends IExplorationExtension {

	/** Returns the expected input element type. */
	Class<I> getInputType();

	/** Returns the output element type. */
	Class<O> getOutputType();

	/**
	 * Flag indicating whether the transformation results of this
	 * {@link IExpSpecTransformationModule} replace input elements, or whether they are an addition.
	 */
	boolean outputIsReplacement();

	/**
	 * Returns a {@link Predicate} that determines which elements shall serve as inputs to the
	 * transformation. If {@code null} is returned, the input {@link ExplorationTarget}s are scanned
	 * for the elements of the input type.
	 */
	Predicate<I> getInputFilter();

	/**
	 * Transforms the {@link ExplorationTarget}s of a DSE run. Examples for this preprocessing
	 * includes a transformation into {@link InternalConstraint}s, or DSML processing.
	 * 
	 * @param mainInjector
	 *            {@link Injector} holding input artifacts and mappers that associate input elements
	 *            with the DSE-internal representation.
	 * @param input
	 *            An input element of the type {@code I} as specified by {@link #getInputType()}.
	 * @return The transformed element replacing the input element.
	 * @throws ExplorationException
	 *             If the transformation fails for any reason.
	 */
	O transform(Injector mainInjector, I input) throws ExplorationException;
}
