/*-------------------------------------------------------------------------+
| Copyright 2015 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.util;

import static org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceConnectionAdapter.ConnectionType.BIDIRECTIONAL;
import static org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceConnectionAdapter.ConnectionType.INCOMING;
import static org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceConnectionAdapter.ConnectionType.OUTGOING;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Map;
import java.util.Optional;
import java.util.function.Predicate;

import org.fortiss.af3.component.model.Channel;
import org.fortiss.af3.component.model.Component;
import org.fortiss.af3.component.model.InputPort;
import org.fortiss.af3.component.model.OutputPort;
import org.fortiss.af3.component.model.Port;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IModelElementAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceConnectionAdapter.ConnectionType;
import org.fortiss.af3.exploration.alg.exception.InvalidModelException;
import org.fortiss.af3.exploration.alg.feature.safetyarch.variable.SafetyFunctionArchEntry;
import org.fortiss.af3.platform.model.ExecutionUnit;
import org.fortiss.af3.platform.model.IPlatformArchitectureElement;
import org.fortiss.af3.platform.model.IPlatformCommunicationResource;
import org.fortiss.af3.platform.model.IPlatformExport;
import org.fortiss.af3.platform.model.IPlatformResource;
import org.fortiss.af3.platform.model.PlatformArchitecture;
import org.fortiss.af3.platform.model.PlatformConnectorUnit;
import org.fortiss.af3.platform.model.TransmissionConnection;
import org.fortiss.af3.safety.model.SafetyArchitecture;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;
import org.fortiss.tooling.base.model.element.IConnection;
import org.fortiss.tooling.base.model.element.IConnector;
import org.fortiss.tooling.base.model.element.IHierarchicElement;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.base.utils.AnnotationUtils;
import org.jgrapht.graph.DefaultDirectedGraph;
import org.jgrapht.graph.DefaultEdge;

/**
 * Utility methods which help to abstract AF3 specifica for the DSE.
 * 
 * @author diewald
 */
public class AF3Utils {

	/** Defines incoming, outgoing, or bidirectional connections by their connoctor types. */
	private static Map<Class<? extends IConnector>, ConnectionType> connectionTypeLUT = Collections
			.unmodifiableMap(new HashMap<Class<? extends IConnector>, ConnectionType>() {/**/
			});

	/** Constructor. Prevent initialization. */
	private AF3Utils() {
	}

	/**
	 * Translates the safety architecture encoded in a {@link SafetyFunctionArchEntry} to a
	 * {@link SafetyArchitecture} enum representation.
	 * 
	 * @param sfaEntry
	 *            {@link SafetyFunctionArchEntry} to be translated.
	 * @return The encoded {@link SafetyArchitecture}, or {@code null} if no matching enum is found.
	 */
	public static SafetyArchitecture getSFArchFromEntry(SafetyFunctionArchEntry sfaEntry) {
		int channelNum = sfaEntry.getNumChannels();
		int triggerNum = sfaEntry.getNumTriggerSignal();
		boolean diagnosisPresent = sfaEntry.isDiagnosicUnitPresent();
		boolean diagnosisConnected = sfaEntry.isDiagnosicUnitConnected();

		if(diagnosisConnected) {
			if(channelNum == 1 && triggerNum == 1) {
				return SafetyArchitecture.ARCH_1OO1_D;
			} else if(channelNum == 2 && triggerNum == 1) {
				return SafetyArchitecture.ARCH_1OO2_D;
			}
		} else {
			if(channelNum == 1 && triggerNum == 1) {
				return diagnosisPresent ? SafetyArchitecture.ARCH_1OO1_DU
						: SafetyArchitecture.ARCH_1OO1;
			} else if(channelNum == 2 && triggerNum == 1) {
				return diagnosisPresent ? SafetyArchitecture.ARCH_1OO2_DU
						: SafetyArchitecture.ARCH_1OO2;
			} else if(channelNum == 3 && triggerNum == 1) {
				return diagnosisPresent ? SafetyArchitecture.ARCH_1OO3_DU
						: SafetyArchitecture.ARCH_1OO3;
			} else if(channelNum == 2 && triggerNum == 2) {
				return diagnosisPresent ? SafetyArchitecture.ARCH_2OO2_DU
						: SafetyArchitecture.ARCH_2OO2;
			} else if(channelNum == 3 && triggerNum == 2) {
				return diagnosisPresent ? SafetyArchitecture.ARCH_2OO3_DU
						: SafetyArchitecture.ARCH_2OO3;
			} else if(channelNum == 4 && triggerNum == 2) {
				return SafetyArchitecture.ARCH_2OO4;
			}
		}

		return null;
	}

	/**
	 * Returns the number of independent safety channels encoded in a {@link SafetyArchitecture}.
	 * 
	 * @param sfArch
	 *            {@link SafetyArchitecture} from which the channel number shall be extracted.
	 * @return Number of channels, or {@code null} if no {@link SafetyArchitecture} is given or the
	 *         architecture is unknown.
	 */
	public static Integer getChannelNumbersSafetyArch(SafetyArchitecture sfArch) {
		if(sfArch == null) {
			return null;
		}

		switch(sfArch) {
			case ARCH_1OO1:
			case ARCH_1OO1_D:
			case ARCH_1OO1_DU:
				return 1;
			case ARCH_1OO2:
			case ARCH_1OO2_D:
			case ARCH_1OO2_DU:
			case ARCH_2OO2:
			case ARCH_2OO2_DU:
				return 2;
			case ARCH_1OO3:
			case ARCH_1OO3_DU:
			case ARCH_2OO3:
			case ARCH_2OO3_DU:
				return 3;
			case ARCH_2OO4:
				return 4;
			default:
				return null;
		}
	}

	/**
	 * Returns the number signals from independent safety channels that are required to trigger the
	 * actual safety function. This information is extracted form the given
	 * {@link SafetyArchitecture}.
	 * 
	 * @param sfArch
	 *            {@link SafetyArchitecture} from which the number of triggers shall be extracted.
	 * @return Number of triggers, or {@code null} if no {@link SafetyArchitecture} is given or the
	 *         architecture is unknown.
	 */
	public static Integer getTriggerNumbersSafetyArch(SafetyArchitecture sfArch) {
		if(sfArch == null) {
			return null;
		}

		switch(sfArch) {
			case ARCH_1OO1:
			case ARCH_1OO1_D:
			case ARCH_1OO1_DU:
			case ARCH_1OO2:
			case ARCH_1OO2_D:
			case ARCH_1OO2_DU:
			case ARCH_1OO3:
			case ARCH_1OO3_DU:
				return 1;
			case ARCH_2OO2:
			case ARCH_2OO2_DU:
			case ARCH_2OO3:
			case ARCH_2OO3_DU:
			case ARCH_2OO4:
				return 2;
			default:
				return null;
		}
	}

	/**
	 * Returns whether the given {@link SafetyArchitecture} encodes the presence of one or more
	 * diagnostic units.
	 * 
	 * @param sfArch
	 *            {@link SafetyArchitecture} encoding the presence of diagnostic unit (s).
	 * @return Presence of diagnostic unit(s), or {@code null} if no {@link SafetyArchitecture} is
	 *         given or the architecture is unknown.
	 */
	public static Optional<Boolean> isDiagPresentSafetyArch(SafetyArchitecture sfArch) {
		switch(sfArch) {
			case ARCH_1OO1_D:
			case ARCH_1OO1_DU:
			case ARCH_1OO2_D:
			case ARCH_2OO2_DU:
			case ARCH_1OO2_DU:
			case ARCH_1OO3_DU:
			case ARCH_2OO3_DU:
				return Optional.of(true);
			case ARCH_1OO1:
			case ARCH_1OO2:
			case ARCH_2OO2:
			case ARCH_1OO3:
			case ARCH_2OO3:
			case ARCH_2OO4:
				return Optional.of(false);
			default:
				return Optional.empty();
		}
	}

	/**
	 * Returns whether the given {@link SafetyArchitecture} encodes the presence of one or more
	 * diagnostic units that directly influence whether the safety function is triggered.
	 * 
	 * @param sfArch
	 *            {@link SafetyArchitecture} encoding the presence of diagnostic unit (s).
	 * @return Presence of diagnostic unit(s) influencing the safety function's output, or
	 *         {@code null} if no {@link SafetyArchitecture} is given or the architecture is
	 *         unknown.
	 */
	public static Optional<Boolean> isDiagConnectedSafetyArch(SafetyArchitecture sfArch) {
		switch(sfArch) {
			case ARCH_1OO1_D:
			case ARCH_1OO2_D:
				return Optional.of(true);
			case ARCH_1OO1:
			case ARCH_1OO1_DU:
			case ARCH_1OO2:
			case ARCH_1OO2_DU:
			case ARCH_2OO2:
			case ARCH_2OO2_DU:
			case ARCH_1OO3:
			case ARCH_1OO3_DU:
			case ARCH_2OO3:
			case ARCH_2OO3_DU:
			case ARCH_2OO4:
				return Optional.of(false);
			default:
				return Optional.empty();
		}
	}

	/**
	 * Determines the {@link ConnectionType} of a connection by examining the class of the given
	 * connector that is the target of the current connection. This is indirectly performed on by a
	 * lookup table.
	 * 
	 * @param sourceConnector
	 *            "sending" connector.
	 * @param targetConnector
	 *            Target {@link IConnector} of the {@link IConnection} whose direction shall be
	 *            determined.
	 * @return {@link ConnectionType} specifying the direction of the {@link IConnection}.
	 */
	public static ConnectionType getConnectionTypeDirection(IConnector sourceConnector,
			IConnector targetConnector) {
		// By the method signature, we know the instance type is a IConnector class.
		ConnectionType connectionTypeSource =
				connectionTypeLUT.get(sourceConnector.eClass().getInstanceClass());
		ConnectionType connectionTypeTarget =
				connectionTypeLUT.get(targetConnector.eClass().getInstanceClass());

		if(connectionTypeSource == OUTGOING && connectionTypeTarget == INCOMING) {
			return OUTGOING;
		} else if(connectionTypeSource == INCOMING && connectionTypeTarget == OUTGOING) {
			return INCOMING;
		} else if((connectionTypeSource == BIDIRECTIONAL || connectionTypeSource == null) &&
				(connectionTypeTarget != null && connectionTypeTarget != BIDIRECTIONAL)) {
			// Revert the direction for the target port, since we are examining the connection from
			// its source.
			return connectionTypeTarget == INCOMING ? OUTGOING : INCOMING;
		} else if((connectionTypeTarget == BIDIRECTIONAL || connectionTypeTarget == null) &&
				(connectionTypeSource != null && connectionTypeSource != BIDIRECTIONAL)) {
			return connectionTypeSource;
		}
		return BIDIRECTIONAL;
	}

	/**
	 * Returns the given {@link IAnnotatedSpecification} type of the given {@link IModelElement} and
	 * throws an exception if the annotation cannot be found.
	 * 
	 * @param element
	 *            {@link IModelElement} to fetch the {@link IAnnotatedSpecification} from.
	 * @param annotationClass
	 *            {@link IAnnotatedSpecification}-Type.
	 * @return The {@link IAnnotatedSpecification} of the given {@link IModelElement}.
	 * @throws InvalidModelException
	 *             if no {@link IAnnotatedSpecification} of the given type exists for the given
	 *             {@link IModelElement}.
	 */
	public static <T extends IAnnotatedSpecification> T getAnnotationChecked(IModelElement element,
			Class<T> annotationClass) throws InvalidModelException {
		T annotation = AnnotationUtils.getAnnotation(element, annotationClass);
		if(annotation == null) {
			throw new InvalidModelException(element, annotationClass);
		}
		return annotation;
	}

	/**
	 * Predicate whether the subject {@link PlatformConnectorUnit} is a connector of the given
	 * {@link IModelElementAdapter}.
	 */
	public static boolean isConnectorOf(PlatformConnectorUnit pConn, IModelElementAdapter element) {
		return pConn.getOwner() == element.getObject(IModelElementAdapter.class);
	}

	/**
	 * Predicate whether the subject {@link PlatformConnectorUnit}'s containing element is contained
	 * in the given {@link IModelElementAdapter} and also contained in the given
	 * {@link DefaultDirectedGraph}.
	 */
	public static boolean isContainedInResource(PlatformConnectorUnit pConn,
			IModelElementAdapter element) {
		return pConn.getOwner() == element.getObject(IHierarchicElement.class).getContainer();
	}

	/**
	 * Predicate whether the subject {@link PlatformConnectorUnit} is the predecessor
	 * {@link PlatformConnectorUnit} for the device-internal communication of the given
	 * {@link IModelElementAdapter}.
	 */
	public static boolean isDeviceInternalPredecessor(PlatformConnectorUnit connCandidate,
			IModelElementAdapter element,
			DefaultDirectedGraph<PlatformConnectorUnit, DefaultEdge> route) {
		// The following rules apply for device internal communication:
		// - Sender Port is on the same device OR
		// - its reference port is one layer "below", e.g. Gateway -> Router AND
		// - the sender connector must have been traversed (natural for path walking a multicast
		// message.
		if(route.containsVertex(connCandidate) && (isConnectorOf(connCandidate, element) ||
				isContainedInResource(connCandidate, element))) {
			return true;
		}
		return false;
	}

	/**
	 * Returns the collection of {@link InputPort}s and {@link OutputPort}s attached to the given
	 * {@link Component}.
	 * 
	 * @param component
	 *            Containing {@link Component}.
	 * @return Collection of attached {@link InputPort}s and {@link OutputPort}s as {@link Port}s.
	 */
	public static Collection<Port> getInputOutputPortsOf(Component component) {
		Collection<Port> ports = new ArrayList<Port>();
		ports.addAll(component.getInputPorts());
		ports.addAll(component.getOutputPorts());
		return ports;
	}

	/**
	 * Returns the collection of {@link IConnector}s which are connected via {@link IConnection}s
	 * with the given {@link IConnector}
	 * 
	 * @param connector
	 *            {@link IConnector} whose connected {@link IConnector}s shall be found.
	 * @return Collection of connected {@link IConnector}s.
	 */
	public static Collection<IConnector> getConnectedPortsOf(IConnector connector) {
		Collection<IConnector> connectedPorts = new ArrayList<>();
		connector.getIncoming().stream().forEach(conn -> connectedPorts.add(conn.getSource()));
		connector.getOutgoing().stream().forEach(conn -> connectedPorts.add(conn.getTarget()));
		return connectedPorts;
	}

	/**
	 * Returns the collection of {@link IConnection}s of the given {@link IConnector} which are
	 * connected to {@link IConnector}s of the target {@link IHierarchicElement}.
	 * 
	 * @param srcConn
	 *            {@link IConnector} whose connected {@link IConnector}s shall be found.
	 * @param tgtElement
	 *            {@link IHierarchicElement} of the {@link IConnection}s.
	 * @return Collection of connecting {@link IConnection}s.
	 */
	public static Map<TransmissionConnection, ConnectionType>
			getConnectionsBetween(PlatformConnectorUnit srcConn, IHierarchicElement tgtElement) {
		Map<TransmissionConnection, ConnectionType> connections = new HashMap<>();
		Predicate<IConnection> sameOwner = c -> c.getSource().getOwner() == tgtElement;
		// TODO(AD, #3722): This casting indicates a "broken" meta-model: getIncoming and
		// getOutgoing should return TransmissionConnections.
		srcConn.getIncoming().stream().filter(sameOwner)
				.forEach(c -> connections.put((TransmissionConnection)c, INCOMING));
		srcConn.getOutgoing().stream().filter(sameOwner)
				.forEach(c -> connections.put((TransmissionConnection)c, OUTGOING));
		return connections;
	}

	/**
	 * Returns the collection of {@link IConnection}s of the given {@link IHierarchicElement} which
	 * are connected to the target {@link IHierarchicElement}.
	 * 
	 * @param srcElement
	 *            whose connections to the {@code tgtElement} shall be found.
	 * @param tgtElement
	 *            {@link IHierarchicElement} of the {@link IConnection}s.
	 * @return Collection of connected {@link Port}s.
	 */
	public static Map<TransmissionConnection, ConnectionType> getConnectionsBetween(
			IPlatformArchitectureElement srcElement, IPlatformArchitectureElement tgtElement) {
		Map<TransmissionConnection, ConnectionType> connections = new HashMap<>();
		srcElement.getPlatformConnectorUnits()
				.forEach(c -> connections.putAll(getConnectionsBetween(c, tgtElement)));
		return connections;
	}

	/**
	 * Returns the collection of <i>all</i> directly connected {@link Channel}s of the given
	 * {@link Component}.
	 * 
	 * @param component
	 *            Subject {@link Component}.
	 * @return Collection of directly connected {@link Channel}s.
	 */
	public static Collection<Channel> getConnectedChannelsOf(Component component) {
		Collection<Port> containedPorts = getInputOutputPortsOf(component);
		Collection<Channel> connectedChannels = new ArrayList<>();
		containedPorts.stream().forEach(
				p -> p.getIncomingChannels().stream().forEach(ch -> connectedChannels.add(ch)));
		containedPorts.stream().forEach(
				p -> p.getOutgoingChannels().stream().forEach(ch -> connectedChannels.add(ch)));
		return connectedChannels;
	}

	/**
	 * <p>
	 * Determines the effective target {@link IConnector}s of a given "source" {@link IConnector}.
	 * Elements/{@link IConnector}s that are irrelevant to communication within the platform are not
	 * added to the target list. Instead, the connectors that are accessible by the source
	 * {@link IConnector} are traversed until relevant targets are identified, like
	 * {@link ExecutionUnit}s selected for the deployment.
	 * </p>
	 * <p>
	 * <b>NOTE:</b> This method may return targets that are not directly involved in the
	 * communication between {@link ExecutionUnit}s. However, 'hull' elements (or similar elements)
	 * that may "disturb" the communication between {@link ExecutionUnit}s will not appear.
	 * </p>
	 * <p>
	 * <b>Implementation Note:</b> The <i>valid</i> targets are determined via path walk following
	 * all connections from the source {@link IConnector} until <i>valid</i> targets are found.
	 * </p>
	 */
	public static Collection<IConnector> getEffectiveConnectionTargets(IConnector sourceConnector,
			Collection<ExecutionUnit> af3DeploymentTargets) {
		Collection<IConnector> targetConnectors = new HashSet<IConnector>();
		if(sourceConnector == null) {
			return targetConnectors;
		}

		LinkedList<IConnector> possibleTargetConnectors = new LinkedList<IConnector>();
		Collection<IConnector> visitedTargetConnectors = new HashSet<IConnector>();

		possibleTargetConnectors.add(sourceConnector);
		while(!possibleTargetConnectors.isEmpty()) {
			IConnector nextTargetConnector = possibleTargetConnectors.poll();

			// This section handles the traversal between the elements of the same platform
			// architecture.
			targetConnectors.addAll(
					seekIntraPlatformTargetConnectors(sourceConnector, possibleTargetConnectors,
							visitedTargetConnectors, nextTargetConnector, af3DeploymentTargets));

			visitedTargetConnectors.add(nextTargetConnector);
		}

		return targetConnectors;
	}

	/**
	 * Identifies those {@link IConnector}s that are actual target {@link IConnector}s of data sent
	 * out by the given {@code sourceConnector}. This methos is needed as in a
	 * {@link PlatformArchitecture}, there exist {@link IConnector}s which are attached to container
	 * elements that are not backed by physical interface ports (These are represented by the
	 * contained elements).
	 * 
	 * @param sourceConnector
	 *            Source {@link IConnector} for which the target {@link IConnector}s need to be
	 *            identified.
	 * @param possibleTargetConnectors
	 *            List to which those {@link IConnector}s are added that are directly connected to
	 *            the {@code sourceConnector} but are no "communication targets" as described above
	 *            an need further traversal.
	 * @param visitedTargetConnectors
	 *            Set of already traversed connectors (to avoid loops).
	 * @param nextTargetConnector
	 *            Current {@link IConnector} whose target {@link IConnector}s need to be searched.
	 * @param deploymentTargets
	 *            Set of execution units present on the target platform.
	 * @return Collection of {@link IConnector}s that are actual target {@link IConnector}s of the
	 *         {@code sourceConnector}.
	 */
	private static Collection<IConnector> seekIntraPlatformTargetConnectors(
			IConnector sourceConnector, LinkedList<IConnector> possibleTargetConnectors,
			Collection<IConnector> visitedTargetConnectors, IConnector nextTargetConnector,
			Collection<ExecutionUnit> deploymentTargets) {
		Collection<IConnector> targetConnectors = new ArrayList<IConnector>();

		Collection<IConnection> attachedConnections = new ArrayList<IConnection>();
		attachedConnections.addAll(nextTargetConnector.getOutgoing());
		attachedConnections.addAll(nextTargetConnector.getIncoming());

		for(IConnection connToTarget : attachedConnections) {
			IConnector connectorTgt = null;
			IConnector connectorSrc = null;
			if(connToTarget.getTarget().getOwner() != nextTargetConnector.getOwner()) {
				connectorTgt = connToTarget.getTarget();
				connectorSrc = connToTarget.getSource();
			}
			if(connToTarget.getSource().getOwner() != nextTargetConnector.getOwner()) {
				connectorTgt = connToTarget.getSource();
				connectorSrc = connToTarget.getTarget();
			}
			if(connectorTgt == null || connectorSrc == null) {
				throw new RuntimeException(
						"Error: \"floating\" connection found during transformation for the DSE.");
			}

			IHierarchicElement tgtElement = connectorTgt.getOwner();
			// Do not traverse back (to same connector) & not to the same IHierarchicElement.
			if(!visitedTargetConnectors.contains(connectorTgt) &&
					tgtElement != sourceConnector.getOwner()) {
				if(tgtElement instanceof IPlatformArchitectureElement && isValidConnectionTarget(
						(IPlatformArchitectureElement)tgtElement, deploymentTargets)) {
					targetConnectors.add(connectorTgt);
				} else if(tgtElement == connectorSrc.getOwner().getContainer() &&
						!(nextTargetConnector instanceof IPlatformExport)) {
					// Treatment of hull ports: Do not walk a path that would just traverse a 'hull'
					// port. Hence, just consider the ports to the "outside" (upper level).
					for(IConnection currConn : connectorTgt.getOutgoing()) {
						IConnector currTgtConnector = currConn.getTarget();
						if(currTgtConnector.getOwner().getContainer() != tgtElement) {
							if(tgtElement instanceof IPlatformArchitectureElement &&
									isValidConnectionTarget(
											(IPlatformArchitectureElement)currTgtConnector
													.getOwner(),
											deploymentTargets)) {
								targetConnectors.add(currTgtConnector);
							} else {
								possibleTargetConnectors.add(currTgtConnector);
							}
						}
					}
					for(IConnection currConn : connectorTgt.getIncoming()) {
						IConnector currTgtConnector = currConn.getSource();
						if(currTgtConnector.getOwner().getContainer() != tgtElement) {
							if(tgtElement instanceof IPlatformArchitectureElement &&
									isValidConnectionTarget(
											(IPlatformArchitectureElement)currTgtConnector
													.getOwner(),
											deploymentTargets)) {
								targetConnectors.add(currTgtConnector);
							} else {
								possibleTargetConnectors.add(currTgtConnector);
							}
						}
					}

					visitedTargetConnectors.add(connectorTgt);
				} else {
					possibleTargetConnectors.add(connectorTgt);
				}
			}
		}

		return targetConnectors;
	}

	/**
	 * Returns whether the given {@link IHierarchicElement} is a valid connection end-point. These
	 * end-points are either {@link ExecutionUnit}s selected as deployment targets for
	 * {@link Component}s or {@link IPlatformCommunicationResource}s whose internal structure is not
	 * modeled.
	 */
	private static boolean isValidConnectionTarget(IPlatformArchitectureElement element,
			Collection<ExecutionUnit> deploymentTargets) {
		if(element instanceof IPlatformResource && deploymentTargets.contains(element)) {
			return true;
		} else if(element instanceof IPlatformCommunicationResource &&
				element.getContainedElements().isEmpty()) {
			return true;
		}
		return false;
	}

	/**
	 * Collect the IConnectors attached to the target elements of the created
	 * IResourceConnectionAdapter for further traversing the PlatformArchitecture.
	 */
	// TODO: Refactor this method to use the static methods from the AF3Utils class instead.
	public static void determineNextConnectionTargets(LinkedList<IConnector> deployableConnectors,
			Collection<IConnector> visitedConnectors, Collection<IHierarchicElement> targetElements,
			Collection<ExecutionUnit> af3DeploymentTargets) {
		for(IHierarchicElement targetElement : targetElements) {
			if(targetElement instanceof IPlatformResource &&
					!af3DeploymentTargets.contains(targetElement)) {
				for(IConnector connectorAtTarget : targetElement.getConnectors()) {
					// Do not traverse back again.
					if(connectorAtTarget != targetElement) {
						if(!visitedConnectors.contains(connectorAtTarget)) {
							deployableConnectors.add(connectorAtTarget);
						}
					}
				}
			}
		}
	}
}
