/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.util;

import static com.google.common.base.Preconditions.checkArgument;
import static java.lang.System.getenv;
import static org.fortiss.tooling.common.util.LambdaUtils.getFirst;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.copyToRefMap;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getChildrenWithType;

import java.lang.annotation.Annotation;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.Set;

import org.apache.commons.lang3.reflect.MethodUtils;
import org.conqat.lib.commons.reflect.ReflectionUtils;
import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.util.EcoreUtil.Copier;
import org.fortiss.af3.component.model.ComponentArchitecture;
import org.fortiss.af3.component.model.OutputPort;
import org.fortiss.af3.exploration.alg.exception.ExplorationException;
import org.fortiss.af3.exploration.alg.exception.ExplorationRuntimeException;
import org.fortiss.af3.exploration.alg.exception.InvalidTimingModelException;
import org.fortiss.af3.exploration.model.ExplorationSpecification;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.fortiss.tooling.kernel.utils.EcoreUtils;

import com.google.common.collect.BiMap;
import com.google.common.collect.Multimap;

/**
 * Static utility methods mainly used in the Design Space Exploration.
 * 
 * @author diewald
 */
public class ExplorationAlgUtils {

	/** Constructor. Prevent initialization. */
	private ExplorationAlgUtils() {
	}

	/** Predicate whether to print verbose debug messages related to encoding validation. */
	public static final boolean DSE_VALIDATE_ENCODING =
			"true".equals(getenv("DSE_VALIDATE_ENCODING"));

	/** Predicate whether to print verbose debug messages related to scheduling. */
	public static final boolean DSE_VALIDATE_SCHEDULE =
			"true".equals(getenv("DSE_VALIDATE_SCHEDULE"));

	/** Predicate whether to print debug messages measuring time. */
	public static final boolean DSE_DEBUG_PRINT_TIME =
			"true".equals(getenv("DSE_DEBUG_PRINT_TIME"));

	/**
	 * Returns a list of source ports within the given {@link ComponentArchitecture}. Source ports
	 * are defined as ports that do not have direct inputs.
	 * 
	 * @param componentArchitecture
	 *            The {@link ComponentArchitecture} for whom source ports shall be identified.
	 * @return {@link EList} of source {@link OutputPort}s.
	 * @throws InvalidTimingModelException
	 *             if the timing specification of any traversed element is inconsistent.
	 */
	public static EList<OutputPort> getSourcePorts(ComponentArchitecture componentArchitecture)
			throws InvalidTimingModelException {
		if(componentArchitecture == null) {
			return null;
		}

		EList<OutputPort> allOutputPortsInArch =
				getChildrenWithType(componentArchitecture, OutputPort.class);
		EList<OutputPort> sourcePortList = new BasicEList<OutputPort>();
		for(OutputPort currentOutputPort : allOutputPortsInArch) {
			// Filter out ports that represent the transition through a hierarchic hull.
			if(currentOutputPort.getIncoming().isEmpty() /*
															 * &&
															 * !isAnyParentComponentDeployable(
															 * currentOutputPort.getComponent())
															 */) {
				// FIXME: HERE
				sourcePortList.add(currentOutputPort);
			}
		}

		return sourcePortList;
	}

	/**
	 * Returns all Classes from the given Collection which inherit from the given class.
	 * 
	 * @param targetClass
	 *            Class providing the type filer.
	 * @param sourceTypeColl
	 *            Collection of Classes to be filtered.
	 * @return Collection of Classes inheriting from the given class.
	 */
	@SuppressWarnings("unchecked")
	public static <T> Collection<Class<T>> pickSubTypesOf(Class<T> targetClass,
			Collection<Class<?>> sourceTypeColl) {
		Collection<Class<T>> filteredColl = new ArrayList<>();
		for(Class<?> srcType : sourceTypeColl) {
			if(targetClass.isAssignableFrom(srcType)) {
				filteredColl.add((Class<T>)srcType);
			}
		}
		return filteredColl;
	}

	/**
	 * Searches for a method annotated with the given {@link Annotation} type in the given class and
	 * returns it. If no such method is found, {@code null} is returned; if more than one annotated
	 * method is found, an exception is thrown.
	 * 
	 * @param containingClass
	 *            Class to be examined.
	 * @param annotationType
	 *            annotation of the seeked method.
	 * @return see above.
	 */
	public static Method getMethodAnnotatedWith(Class<?> containingClass,
			Class<? extends Annotation> annotationType) {
		// From Google Guice --> Matchers.
		Retention retention = annotationType.getAnnotation(Retention.class);
		checkArgument(retention != null && retention.value() == RetentionPolicy.RUNTIME,
				"Annotation %s is missing RUNTIME retention", annotationType.getSimpleName());
		// END Matchers snippet.
		Method annotatedMethod = null;
		for(Method method : containingClass.getMethods()) {
			if(method.getAnnotationsByType(annotationType).length != 0) {
				if(annotatedMethod == null) {
					// First match.
					annotatedMethod = method;
				} else {
					// Multiple methods with the given annotation found.
					throw new ExplorationRuntimeException("Found multiple methods in the class " +
							containingClass.getSimpleName() + " annotated with " +
							annotationType.getSimpleName() + ".");
				}
			}
		}
		return annotatedMethod;
	}

	/**
	 * Returns the typed key from the map which is the best match for the given
	 * class. The best match is defined by the first match occurring in a breath
	 * first search of the inheritance tree, where the base class is always
	 * visited before the implemented interfaces. Interfaces are traversed in
	 * the order they are defined in the source file. The only exception is {@link Object}, which is
	 * considered only as the very last option.
	 * <p>
	 * As this lookup can be expensive (reflective iteration over the entire inheritance tree) the
	 * results should be cached if multiple lookups for the same class are expected.
	 * 
	 * @param clazz
	 *            the class being looked up.
	 * @param keySet
	 *            the set of keys to perform the lookup in.
	 * @return the best match found or <code>null</code> if no matching key
	 *         was found. Note that <code>null</code> will also be returned if
	 *         the entry for the best matching class was <code>null</code>.
	 * 
	 *         Adapted from {@link ReflectionUtils#performNearestClassLookup(Class, Map)}.
	 */
	@SuppressWarnings("unchecked")
	public static <K> Class<? extends K> performNearestClassLookup(Class<? extends K> clazz,
			Set<Class<? extends K>> keySet) {
		Queue<Class<? extends K>> q = new LinkedList<Class<? extends K>>();
		q.add(clazz);

		Class<? extends K> current = null;
		while(!q.isEmpty()) {
			current = q.poll();
			if(keySet.contains(current)) {
				return current;
			}

			Class<?> superClass = current.getSuperclass();
			if(superClass != null && superClass != Object.class) {
				q.add((Class<? extends K>)superClass);
			}

			for(Class<?> iface : current.getInterfaces()) {
				q.add((Class<? extends K>)iface);
			}
		}
		return current;
	}

	/**
	 * Returns the value from the {@link Multimap}, whose key is the best match for the given
	 * class. The best match is defined by the first match occurring in a breath
	 * first search of the inheritance tree, where the base class is always
	 * visited before the implemented interfaces. Interfaces are traversed in
	 * the order they are defined in the source file. The only exception is {@link Object}, which is
	 * considered only as the very last option.
	 * <p>
	 * As this lookup can be expensive (reflective iteration over the entire inheritance tree) the
	 * results should be cached if multiple lookups for the same class are expected.
	 * 
	 * 
	 * @param clazz
	 *            the class being looked up.
	 * @param classMap
	 *            the {@link Multimap} to perform the lookup in.
	 * @return the best match found or the empty set if no matching entry
	 *         was found. Note that the empty set will also be returned if
	 *         the entry for the best matching class was the empty set.
	 * 
	 *         Adapted from {@link ReflectionUtils#performNearestClassLookup(Class, Map)}.
	 */
	public static <K, T> Collection<T> performNearestClassLookup(Class<? extends K> clazz,
			Multimap<Class<? extends K>, T> classMap) {
		Class<? extends K> key = performNearestClassLookup(clazz, classMap.keySet());
		return classMap.get(key);
	}

	/**
	 * Returns a {@link Method} of the given class annotated with the given {@link Annotation}.
	 * If not <i>exactly</i> ONE method is specified with the given annotation, an exception is
	 * thrown. {@code doThrow} is used, a {@link RuntimeException} is thrown instead
	 * of just returning {@code false}.
	 * 
	 * @param classObj
	 *            class to check.
	 * @param annotation
	 *            {@link Annotation} to find.
	 * @return method annotated with the given {@link Annotation}.
	 * 
	 * @throws ExplorationException
	 *             if none or more {@link Method}s are annotated with the given {@link Annotation}.
	 */
	public static Method getAnnotatedMethod(Class<?> classObj,
			Class<? extends Annotation> annotation) throws ExplorationException {
		// The desired result is specified by the evaluator name and the return type.
		List<Method> method = MethodUtils.getMethodsListWithAnnotation(classObj, annotation);
		if(method.size() != 1) {
			throw new ExplorationException("The class " + classObj.toString() +
					" must contain exactly one method annotated with " + annotation.toString() +
					".");
		}
		// Existence ensured by the above check.
		return getFirst(method).get();
	}

	/**
	 * Returns a self-contained copy of the given <code>eObject</code>.
	 * Here, the copier that performs the copy operation must be given. The copier (which is a
	 * LinkedHashMap) can be further utilized to obtain information about the copied elements.
	 * 
	 * @param eObject
	 *            the object to copy.
	 * @param copier
	 *            This copier will be used to perform the actual copy operation.
	 * @return the copy.
	 * @see Copier
	 */
	public static <T extends EObject> T copy(T eObject, Copier copier) {
		EObject result = copier.copy(eObject);
		copier.copyReferences();

		@SuppressWarnings("unchecked") T t = (T)result;
		return t;
	}

	/**
	 * Returns a copy of the given {@link ExplorationSpecification} and its referenced
	 * {@link ExplorationTarget}s. Additionally, the EMF notifications are disabled for these
	 * elements to ease manipulation of the data structure.
	 * 
	 * @param origExpSpec
	 *            {@link ExplorationSpecification} to be copied.
	 * @param copyRefMap
	 *            Association map of the original {@link ExplorationTarget}s to their respective
	 *            copies.
	 * @return Copy of the input {@link ExplorationSpecification} with disabled notifications.
	 */
	public static ExplorationSpecification copyExplorationSpecification(
			ExplorationSpecification origExpSpec, BiMap<EObject, EObject> copyRefMap) {
		ExplorationSpecification expSpec = EcoreUtils.copy(origExpSpec);
		expSpec.getTargets().clear();

		origExpSpec.getTargets().forEach(origTgt -> {
			Map<EObject, EObject> refMap = copyToRefMap(origTgt);
			ExplorationTarget<?> expTarget = (ExplorationTarget<?>)refMap.get(origTgt);
			expSpec.getTargets().add(expTarget);
			copyRefMap.putAll(refMap);
		});

		return expSpec;
	}
}
